from algorithms.cloud.ms.object_detection import ObjectDetectionMS
from algorithms.cloud.ms.scene_text_detection import SceneTextDetectionMS
from algorithms.cloud.ms.captioning import CaptioningMS
from algorithms.cloud.ms.tagging import TaggingMS
from algorithms.cloud.cloud_preprocessor import CloudPreprocessor

from algorithms.general.visualize import Visualize
from algorithms.general.text_to_speech import TextToSpeech
from algorithms.general.garbage_collect import GarbageCollector
from algorithms.general.utils import AutoIdle, LimitedQ
from algorithms.general.internet_checker import InternetChecker
from algorithms.general.auth import AuthProvider

from algorithms.obstacle_avoidance.obstacle_avoidance import ObsatcleAvoidance
from algorithms.map_path_following.path_tracker import PathTracker
from algorithms.object_tracking.object_tracker import ObjectTracker
from algorithms.image_path_following.match_propeller import MatchPropeller

from algorithms.constants import (
    CLOUD_IMAGE_WIDTH,
    CLOUD_IMAGE_HEIGHT,
    MS_ENDPOINT,
    OBJ_DET_TIME_INTERVAL,
    IMG_CAPTION_TIME_INTERVAL,
    IMG_TAG_TIME_INTERVAL,
    TEXT_TO_SPEECH_TIME_INTERVAL,
    SECONDARY_FRAME_RATE,
)
from algorithms.config import Config

def get_algorithms(root, config:Config, logger, auth_provider):
    obstacle_avoidance = ObsatcleAvoidance(root, config, logger)
    objectr_tracker = ObjectTracker(root, config, logger)
    match_tracker = MatchPropeller(root, config, logger, auth_provider)
    path_tracker = PathTracker(root, config, logger, auth_provider)
    return obstacle_avoidance, objectr_tracker, match_tracker, path_tracker

def get_ms_engines(config:Config, logger):
    cloud_image_size = (CLOUD_IMAGE_WIDTH, CLOUD_IMAGE_HEIGHT)
    scene_text_spotter = SceneTextDetectionMS(
        config.cloud.scene_text_spotting,
        MS_ENDPOINT,
        logger,
        OBJ_DET_TIME_INTERVAL,
        cloud_image_size,
    )
    image_captioner = CaptioningMS(
        config.cloud.image_captioning,
        MS_ENDPOINT,
        logger,
        IMG_CAPTION_TIME_INTERVAL,
        cloud_image_size,
    )
    image_tagger = TaggingMS(
        config.cloud.image_tagging,
        MS_ENDPOINT,
        logger,
        IMG_TAG_TIME_INTERVAL,
        cloud_image_size,
    )
    cloud_preprocessor = CloudPreprocessor(config)
    return scene_text_spotter, image_captioner, image_tagger, cloud_preprocessor


def get_io_and_utils(config:Config, logger):
    visualize = Visualize(config)
    text_to_speech = TextToSpeech(config)
    garbage_collector = GarbageCollector()
    auto_idle = AutoIdle()
    internet_checker = InternetChecker(logger)
    auth_provider = AuthProvider(logger)
    fps_er = LimitedQ(50)
    return (
        visualize,
        text_to_speech,
        garbage_collector,
        auto_idle,
        internet_checker,
        auth_provider,
        fps_er,
    )
