import threading
import time
import numpy as np
import time
from algorithms.constants import IDLE_THRESHOLD
import rclpy
from rclpy.clock import Clock
from rclpy.time import Time
from rclpy.duration import Duration
class AverageMeter:
    """Computes and stores the average and current value"""
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

def tryconvert(value, default, *types):
    for t in types:
        try:
            return t(value)
        except (ValueError, TypeError):
            continue
    return default


class LimitedQ(list):
    def __init__(self, size ):
        self.size = size
    def append(self, item):
        list.append(self, item)
        if len(self) > self.size: del self[0]
    @property
    def avg(self):
        return self.mean()
    def mean(self):
        if len(self) == 0:
            return 0
        return sum(self) / len(self)
    
class PeriodicThread:
    def __init__(self, interval, func, *args, **kwargs):
        self.interval = interval
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.running = False
        self.thread = None

    def start(self):
        if not self.running:
            self.running = True
            self.thread = threading.Thread(target=self.run)
            self.thread.start()

    def stop(self):
        if self.running:
            self.running = False
            self.thread.join()

    def run(self):
        while self.running:
            start_time = Clock().now()
            self.func(*self.args, **self.kwargs)
            elapsed_time =  (Clock().now() - start_time).nanoseconds / 1e9
            sleep_time = self.interval - elapsed_time
            if sleep_time > 0:
                time.sleep(sleep_time)

class AutoIdle:
    def __init__(self):
        self.last_n_walking = LimitedQ(5)
        self.last_n_spinning = LimitedQ(5)
        self.walking_value = 0.0
        self.spinning_value = 0.0
        self.last_move = time.time()
        self.last_imu_update = time.time()
        self.last_frame_update = time.time()

    def update_frame_time(self):
        self.last_frame_update = time.time()

    def update_imu(self, acc, gyro):
        self.last_n_walking.append(np.sqrt(np.sum(np.array(acc)**2)))
        self.last_n_spinning.append(np.sqrt(np.sum(np.array(gyro)**2)))
        if len(self.last_n_walking) > 1:
            self.walking_value = np.abs(np.diff(self.last_n_walking)).mean()
            self.spinning_value = np.abs(np.diff(self.last_n_spinning)).mean()
        if self.walking_value > IDLE_THRESHOLD or self.spinning_value > IDLE_THRESHOLD:
            self.last_move = time.time()
        self.last_imu_update = time.time()

    def idling_time(self):
        imu_idle_time = time.time() - self.last_move
        return imu_idle_time
    
    def get_motion_data(self):
        return self.walking_value, self.spinning_value
    
    def is_idle(self, threshold=5):
        imu_enabled = time.time() - self.last_imu_update < 2
        camera_enabled = time.time() - self.last_frame_update < 2
        imu_idling = self.idling_time() > threshold
        # ie ce ii
        # t t t -> True
        # t t f -> False
        # t f t -> True
        # t f f -> True
        # f t t -> False
        # f t f -> False
        # f f t -> True
        # f f f -> True        
        if imu_enabled and camera_enabled and imu_idling:
            return True
        elif imu_enabled and camera_enabled and not imu_idling:
            return False
        elif imu_enabled and not camera_enabled:
            return True
        elif not imu_enabled and camera_enabled:
            return False
        else:
            return True


class LogTimeDebugger:
    def __init__(self, name):
        self.name = name
        self.prev_time = time.time()
        self.count = 0

    def log(self):
        print(f"{self.name} {self.count} {time.time() - self.prev_time}")
        self.count += 1
        self.prev_time = time.time()

def get_bool_from_txt(path):
    with open(path, "r") as f:
        line = f.readline().strip()
        f.close()
    return bool(int(line))

def write_txt(path,value=None):
    with open(path, "w") as f:
        if value:
            f.write(str(value))
        f.close()
