import cv2
import datetime
import numpy as np
from algorithms.constants import (
    SENSOR_MAP,
    WIDTH,
    HEIGHT,
    PATH_OBJECTS,
    SHOW_HEIGHT,
    SHOW_WIDTH,
)
from algorithms.obstacle_avoidance.obstacle_detect import ObstacleDetect
from algorithms.config import Config
from algorithms.general.utils import PeriodicThread
CONVERT_RATIO = (SHOW_WIDTH / WIDTH, SHOW_HEIGHT / HEIGHT)
def coord_convert(x,y):
    ratio_w, ratio_h = CONVERT_RATIO
    return  int(x * ratio_w), int(y * ratio_h)


class Visualize:
    def __init__(self, config:Config):
        self.config = config
        self.reconfigure(config)
        self.width = SHOW_WIDTH 

        self.text_font = cv2.FONT_HERSHEY_SIMPLEX
        self.text_size = 0.5 *CONVERT_RATIO[0]
        self.text_thickness =int( 1 * CONVERT_RATIO[0]) 

        self.icre = 10 
        self.fsize = 0.25  * CONVERT_RATIO[0]
        self.bg_text_size = 0.3  * CONVERT_RATIO[0]
                    
        self.drawing_thread = PeriodicThread(0.05, self.draw)
        self.ann0, self.ann1 = None, None
        self.is_idle = False
        self.draw_bag = None
    def start(self):
        if not self.drawing_thread.running:
            self.drawing_thread.start()
    def stop(self):
        if self.drawing_thread.running:
            self.drawing_thread.stop()
    def release(self):
        self.stop()
            
    def reconfigure(self, config:Config):
        self.config = config
        self.height = SHOW_HEIGHT
        if self.config.image_path_following.enable:
            self.height = int(self.height * 1.5)

    def update(
        self,
        is_idle,
        frame_raw0,
        frame_raw1,
        fps,
        contours,
        apf_route,
        goal,
        obs_turning_point,
        obs_dir,
        obstacle_detect,
        matcher_states,
        sensor_data,
        loop_vis,
        object_detected,
        text_detected,
        image_caption,
        image_tags,
        debug_info,
    ):
        self.is_idle = is_idle
        self.draw_bag = dict(
            frame_raw0= frame_raw0 if frame_raw0 is not None else None,
            frame_raw1=frame_raw1 if frame_raw1 is not None else None,
            fps=fps,
            contours=contours,
            apf_route=apf_route,
            goal=goal,
            obs_turning_point=obs_turning_point,
            obs_dir=obs_dir,
            obstacle_detect=obstacle_detect,
            matcher_states=matcher_states,
            sensor_data=sensor_data,
            loop_vis=loop_vis,
            object_detected=object_detected,
            text_detected=text_detected,
            image_caption=image_caption,
            image_tags=image_tags,
            debug_info=debug_info,
        )
    
    def get_annotations(self):
        return self.ann0, self.ann1
        
        
    def draw(self):
        if not self.config.visualize.enable:
            return
        if self.is_idle:
            return
        if self.draw_bag is None:
            return
        if self.draw_bag["frame_raw0"] is None:
            return
        ############################## draw bag release ##############################
        frame_raw0 = self.draw_bag["frame_raw0"]
        frame_raw1 = self.draw_bag["frame_raw1"]
        fps = self.draw_bag["fps"]
        contours = self.draw_bag["contours"]
        apf_route = self.draw_bag["apf_route"]
        goal = self.draw_bag["goal"]
        obs_turning_point = self.draw_bag["obs_turning_point"]
        obs_dir = self.draw_bag["obs_dir"]
        obstacle_detect = self.draw_bag["obstacle_detect"]
        matcher_states = self.draw_bag["matcher_states"]
        sensor_data = self.draw_bag["sensor_data"]
        loop_vis = self.draw_bag["loop_vis"]
        object_detected = self.draw_bag["object_detected"]
        text_detected = self.draw_bag["text_detected"]
        image_caption = self.draw_bag["image_caption"]
        image_tags = self.draw_bag["image_tags"]
        debug_info = self.draw_bag["debug_info"]
        
        ############################## first camera ##############################
        image = self.process_frame_uint8(frame_raw0)
        # image = frame_raw0
        self.put_bounded_text(
            image,
            "FPS: {:.2f}".format(fps),
            coord_convert(2 , 15),
        )
        self.put_bounded_text(
            image,
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3],
            coord_convert(205, 10),
            text_size=self.fsize,
        )
        ############################## debug ##############################
        if self.config.visualize.debug:
            if sensor_data is not None:
                self.put_bounded_text(
                    image,
                    f'az: {sensor_data[SENSOR_MAP["az"]]:.2f} gz: {sensor_data[SENSOR_MAP["gz"]]:.2f} yaw: {sensor_data[SENSOR_MAP["yaw"]]:.2f}',
                    coord_convert(205, 20),
                    text_size=self.fsize,
                )
            if len(loop_vis) > 0:
                self.put_bounded_text(
                    image,
                    f"Total Time: {sum(loop_vis): 4f}",
                    coord_convert(100, 6),
                    text_size=self.fsize,
                )
                image = self.hori_bar_with_list_as_proportion(
                    image, loop_vis, coord_convert(100, 8), *coord_convert(100, 6),
                )
            image = self.draw_debug_info(image, debug_info)


        ############################## obs ##############################
        image = self.draw_contours(image, contours)
        image = self.draw_route(image, apf_route, goal, obs_turning_point, obs_dir)        
        if self.config.obstacle_avoidance.use_obstacle_detect:
            image = self.draw_obstacle(image, obstacle_detect)

        ############################## matcher ##############################
        image = self.draw_matchings(image, matcher_states)
        
        ############################## second camera ##############################
        if frame_raw1 is not None:
            detect_frame = self.process_frame_uint8(frame_raw1)
        else:
            detect_frame = image
        detect_image = self.draw_detection_boxes(detect_frame, object_detected)
        detect_image = self.draw_text_boxes(detect_image, text_detected)
        detect_image = self.draw_captions(detect_image, image_caption)
        detect_image = self.draw_tags(detect_image, image_tags)
        
        self.ann0 = image.copy()
        self.ann1 = detect_image.copy()
            
    def draw_contours(self, image, contours):
        if contours is not None:
            filled_image = np.zeros_like(image)
            for contour in contours:
                try:
                    contour = [[list(coord_convert(x, y)) for x, y in a] for a in contour]
                    contour = np.array(contour).astype(np.int32)
                    image = cv2.drawContours(image, [contour], -1, (0, 255, 0), 1)
                    cv2.fillPoly(filled_image, [contour], (0, 255, 0, 255))
                except:
                    continue
            image = cv2.addWeighted(image, 1, filled_image, 0.2, 0)
        return image

    def draw_route(self, image,apf_route, goal, turning_point, turn_direction):
        if apf_route is not None:
            route_color = (255, 0, 0)
            apf_route = [[list(coord_convert(x, y)) for x, y in a] for a in apf_route]
            apf_route = np.array(apf_route).astype(np.int32)
            if apf_route.shape[0] == 2:
                image = cv2.arrowedLine(
                    image,
                    (apf_route[0][0][0], apf_route[0][0][1]),
                    (apf_route[1][0][0], apf_route[1][0][1]),
                    route_color,
                    2,
                    tipLength=0.2,
                )
            else:
                image = cv2.polylines(image, [apf_route], False, route_color, 2)

        # draw goal
        if goal is not None:
            goal = coord_convert(goal[0], goal[1])
            image = cv2.circle(
                image,
                (goal[0], goal[1]), #this is not oxymoron, dont change to 'goal'
                radius=4,
                color=(0, 255, 255),
                thickness=-10,
            )
            
        if turning_point is not None and turn_direction is not None:
            x, y = coord_convert(turning_point[0], turning_point[1])
            text_size =self.text_size

            image = cv2.circle(
                image, (x, y), radius=4, color=(255, 255, 255), thickness=-10
            )
            # Turning left
            if turn_direction < 0:
                turn_text =f"<<:{-turn_direction :.2f}"
            # Turning right
            elif turn_direction > 0:
                turn_text = f">>:{-turn_direction :.2f}"
            # Going straight
            else:
                turn_text = "^^:--"
            self.put_bounded_text(
                image,
                turn_text,
                coord_convert (2, int(HEIGHT * 0.97)),
                text_size=text_size,
            )
        return image

    def draw_matchings(self, image, matcher_states):
        if matcher_states is None: return image
        if len(matcher_states["match_box"]) == 4:
            self.put_bounded_text(
                image,
                "Matched: Go",
                coord_convert(2, int(HEIGHT * 0.8)),
            )

        if (
            matcher_states["matching_img"] is not None
            and self.config.image_path_following.enable
        ):
            matching_img = cv2.resize(
                matcher_states["matching_img"],
                (int(SHOW_WIDTH * 2), SHOW_HEIGHT),
                cv2.INTER_LINEAR,
            )
            image = self.vertical_stack(image, matching_img)
        return image

    def draw_debug_info(self, image, map):
        basex, basey = 2, 25
        texts, positions = [], []
        for i in map:
            texts.append(f"{i}: {map[i] :6f}" if type(map[i]) == float else f"{i}: {map[i]}")
            positions.append(coord_convert(basex, basey))
            basey += self.icre
        self.put_bounded_texts(
            image,
            texts,
            positions,
            text_size=self.fsize,
            opacity=0.6,
        )
        return image

    def draw_detection_boxes(self, image, result):
        for box in result:
            label = f"{box['cls']}, id: {box['id']}"
            # normalize the box to the input image size
            xy1 = coord_convert(int(box["left"] ), int(box["top"]))
            xy2 = coord_convert(int(box["right"]), int(box["bottom"] ))
            if box["cls"] in PATH_OBJECTS:
                image = cv2.rectangle(image, xy1, xy2, (0, 225, 225), 2)
                self.put_background_text(image, label, coord_convert(box["left"], box["top"]+4), bgcolor=(0, 225, 225))
            # else:
            #     image = cv2.rectangle(image, xy1, xy2,(0, 0, 255), 1)
            #     self.put_background_text(image, label, xy1, bgcolor= (0,0,225))
        return image

    def draw_text_boxes(self, image, boxes):
        h, w = image.shape[:2]
        for box in boxes:
            box_cords = box["box"].copy()
            # convert the normalized pts with the hw
            box_cords[:, :, 0] = box_cords[:, :, 0] * h
            box_cords[:, :, 1] = box_cords[:, :, 1] * w
            box_cords = box_cords.astype(np.int32)
            x1, y1 = box_cords[0][0]
            self.put_background_text(
                image, f"{box['text']}", (x1, y1), bgcolor=(0, 225, 0)
            )
            image = cv2.polylines(
                image, [box_cords], isClosed=True, color=(0, 255, 0), thickness=1
            )
        return image

    def draw_captions(self, image, captions):
        for caption in captions:
            text = caption["text"]
            conf = caption["confidence"]
            self.put_bounded_text(
                image,
                f"{text}",
                coord_convert(int(0.25 * WIDTH), HEIGHT - 10),
                text_size=self.fsize,
            )
        return image

    def draw_tags(self, image, tags):
        for i, tag in enumerate(tags):
            text = tag["name"]
            conf = tag["confidence"]
            self.put_bounded_text(
                image,
                f"{text}",
                coord_convert(int(0.85 * WIDTH), HEIGHT - 20 - i * 10),
                text_size=self.fsize,
            )
        return image

    def put_bounded_text(
        self,
        image,
        text,
        position,
        text_font=None,
        text_size=None,
        text_thickness=None,
        line_type=None,
        opacity=1,
    ):
        text_font = self.text_font if text_font is None else text_font
        text_size = self.text_size if text_size is None else text_size
        text_thickness = (
            self.text_thickness if text_thickness is None else text_thickness
        )
        line_type = cv2.LINE_AA if line_type is None else line_type
        if opacity !=1:
            overlay = image.copy() # must copy()
        else:
            overlay = image
        cv2.putText(
            overlay,
            text,
            position,
            text_font,
            text_size,
            (0, 0, 0),
            int(text_thickness *2),
            line_type,
        )
        cv2.putText(
            overlay,
            text,
            position,
            text_font,
            text_size,
            (255, 255, 255),
            text_thickness,
            line_type,
        )
        # cv2.addWeighted(image, 1-opacity, overlay, opacity, 0, image)
        if opacity !=1:
            cv2.addWeighted(image, 1-opacity, overlay, opacity, 0, image)
            
    def put_bounded_texts(
        self,
        image,
        texts,
        positions,
        text_font=None,
        text_size=None,
        text_thickness=None,
        line_type=None,
        opacity=1,
    ):
        text_font = self.text_font if text_font is None else text_font
        text_size = self.text_size if text_size is None else text_size
        text_thickness = (
            self.text_thickness if text_thickness is None else text_thickness
        )
        line_type = cv2.LINE_AA if line_type is None else line_type
        if opacity !=1:
            overlay = image.copy()
        else:
            overlay = image
        for text, position in zip(texts, positions):
            cv2.putText(
                overlay,
                text,
                position,
                text_font,
                text_size,
                (0, 0, 0),
                int(text_thickness *2),
                line_type,
            )
            cv2.putText(
                overlay,
                text,
                position,
                text_font,
                text_size,
                (255, 255, 255),
                text_thickness,
                line_type,
            )
        if opacity !=1:
            cv2.addWeighted(image, 1-opacity, overlay, opacity, 0, image)
        
    def put_background_text(
        self,
        image,
        text,
        position,
        text_font=None,
        text_size=None,
        text_thickness=None,
        line_type=None,
        color=(0, 0, 0),
        bgcolor=(255, 255, 255),
    ):
        cv2.UMat(image)
        text_font = self.text_font if text_font is None else text_font
        text_size = self.bg_text_size if text_size is None else text_size
        text_thickness = (
            self.text_thickness if text_thickness is None else text_thickness
        )
        line_type = cv2.LINE_AA if line_type is None else line_type
        size = cv2.getTextSize(text, text_font, text_size, text_thickness)
        x, y = position
        w, h = size[0]
        cv2.rectangle(image, (x, y), (x + w, y - h), bgcolor, -1)
        cv2.putText(image, text, position, text_font, text_size, color, text_thickness)

    @staticmethod
    def draw_obstacle(image, obstacle_detect:ObstacleDetect):
        if obstacle_detect.left_tangent.valid:
            image = cv2.line(
                image,
                coord_convert(*obstacle_detect.left_tangent.get_line()[0]),
                coord_convert(*obstacle_detect.left_tangent.get_line()[1]),
                (255, 125, 0),
                1,
            )
        if obstacle_detect.right_tangent.valid:
            image = cv2.line(
                image,
                coord_convert(*obstacle_detect.right_tangent.get_line()[0]),
                coord_convert(*obstacle_detect.right_tangent.get_line()[1]),
                (255, 125, 0),
                1,
            )
        if obstacle_detect.hori_tangent.valid:
            image = cv2.line(
                image,
                coord_convert(*obstacle_detect.hori_tangent.get_line()[0]),
                coord_convert(*obstacle_detect.hori_tangent.get_line()[1]),
                (255, 125, 0),
                1,
            )
        return image

    @staticmethod
    def hori_bar_with_list_as_proportion(image, vis, pos, width, height):
        color1, color2 = (0, 0, 255), (0, 255, 0)
        total = sum(vis)
        if total == 0:
            return image
        x, y = pos
        for i in range(len(vis)):
            w = int(width * vis[i] / total)
            color = color1 if i % 2 == 0 else color2
            # put number as well
            image = cv2.putText(
                image,
                f'{i}',
                (x, y +40),
                cv2.FONT_HERSHEY_SIMPLEX,
                1,
                color,
                2,
                cv2.LINE_AA,
            )
            image = cv2.rectangle(image, (x, y), (x + w, y + height), color, -1)
            x += w
        return image

    @staticmethod
    def horizontal_stack(image, image2):
        if image2 is None:
            return None
        width = image.shape[1] + image2.shape[1]
        height = max(image.shape[0], image2.shape[0])
        showimg = np.zeros((height, width, 3), dtype=np.uint8)
        showimg[0 : image.shape[0], 0 : image.shape[1]] = image
        showimg[
            0 : image2.shape[0], image.shape[1] : image.shape[1] + image2.shape[1]
        ] = image2
        return showimg

    @staticmethod
    def vertical_stack(img1, img2, img2_factor=0.5):
        if img2 is None:
            return img1

        # Get the width and height of the images
        h1, w1 = img1.shape[:2]
        h2, w2 = img2.shape[:2]

        img2 = cv2.resize(img2, (int(w2 * img2_factor), int(h2 * img2_factor)))
        h2, w2 = img2.shape[:2]

        # Add zero padding to the second image
        if w2 < w1:
            pad_width = w1 - w2
            img2 = cv2.copyMakeBorder(
                img2, 0, 0, 0, pad_width, cv2.BORDER_CONSTANT, value=0
            )
        elif w2 > w1:
            pad_width = w2 - w1
            img1 = cv2.copyMakeBorder(
                img1, 0, 0, 0, pad_width, cv2.BORDER_CONSTANT, value=0
            )

        stacked_img = cv2.vconcat([img1, img2])
        return stacked_img

    @staticmethod
    def process_frame_uint8(im, size=(SHOW_WIDTH, SHOW_HEIGHT)):
        # image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = cv2.resize(im, size, cv2.INTER_LINEAR) if size is not None else im
        # image = cv2.normalize(image, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_32F)
        return image
