import cv2
import time
import numpy as np
from algorithms.constants import (
    IMAGE_MATCHING_HEIGHT,IMAGE_MATCHING_WIDTH,
    WIDTH,HEIGHT,
    IMAGE_MATCHING_CHANGING_THRESHOLD,IMAGE_MATCHING_EXPLORATION_THRESHOLD)
from algorithms.general.utils import LimitedQ
from .utils import *

TEMPLATE_CHECK_RANGE = 5 
# MATCH_THRESHOLD = 7
MATCH_THRESHOLD = 5
WINDOW_BARRIER = MATCH_THRESHOLD/3.5

# ORB
NFEATURES = 1000
SCALE = 1.9
NLEVLES = 8

# Matcher
TABLE_NUMBER = 6
KEY_SIZE = 12
MULTI_LEVEL =2
CHECKS = 50
SIM = 0.93
KP_DIST_THRESHOLD = 0.5 * WIDTH # (in pixels)

LEN_N_MATCHES = 3

SKIP_STEP=3


class Template:
    def __init__(self, id, image, cat=None):
        self.id = id
        self.image = image
        self.image =  cv2.resize(image, (IMAGE_MATCHING_WIDTH, IMAGE_MATCHING_HEIGHT)) if image is not None else None
        self.cat = cat

class TemplateInfo:
    def __init__(self, Template, kp, des):
        self.template = Template
        self.n_matches = LimitedQ(LEN_N_MATCHES)
        self.kp = kp
        self.des = des

    def clear(self):
        self.n_matches = LimitedQ(LEN_N_MATCHES)

draw_params = dict(matchColor = (0,255,0),
            singlePointColor = (255,0,0),
            flags = cv2.DrawMatchesFlags_DEFAULT)

class BidirectionalImageMatcher:
    def __init__(self,model):
        # image features
        self.image = None
        self.kp = None
        self.des = None

        # template features
        self.all_template_info = []

        # template range to match
        self.radius = TEMPLATE_CHECK_RANGE

        # template matching parameters
        self.match_threshold = MATCH_THRESHOLD
        self.sim = SIM
        self.kp_dist_threshold = KP_DIST_THRESHOLD
        
        if model != None:
            # using the SuperPoint ML model
            self.kp_detector = model
        else:
            self.kp_detector= cv2.ORB_create(NFEATURES,  SCALE, nlevels=NLEVLES,)
        self.matcher = cv2.FlannBasedMatcher(
            dict(algorithm = 6,
                table_number = TABLE_NUMBER, # 12
                key_size = KEY_SIZE,     # 20
                multi_probe_level = MULTI_LEVEL) ,
            dict(checks=CHECKS))

        self.walking = False
        self.spinning = False
        self.is_processing = False
        
        self.reset()

    def reset(self):
        self.first_match = False
        self.match_box=[]
        self.micro_adjust_dir = LimitedQ(3)
        self.likely_template_idx = 0
        self.n_matches=0

        self.last_change_template_time = time.time()
        self.matching_img = np.zeros((HEIGHT,WIDTH*2,3), np.uint8)
        self.detected_objects = []

    def forward(self):
        self.likely_template_idx = min(self.likely_template_idx + SKIP_STEP, len(self.all_template_info)-1)
        self.last_change_template_time = time.time()
        self.clear_nmatch()

    def backward(self):
        self.likely_template_idx = max(self.likely_template_idx - SKIP_STEP, 0)        
        self.last_change_template_time = time.time()
        self.clear_nmatch()

    def set_state(self, walking, spinning):
        self.walking = walking
        self.spinning = spinning

    def set_image(self, im):
        self.image = cv2.resize(im, (IMAGE_MATCHING_WIDTH, IMAGE_MATCHING_HEIGHT))
        self.image = cv2.cvtColor(self.image, cv2.COLOR_BGR2GRAY)

    def set_detected_objects(self, objects):
        self.detected_objects = objects

    def set_all_template(self, all_templates):
        self.all_template_info=[]
        for template in all_templates:
            template = Template(
                template['image_id'],
                template['image'],
                template['category'],)
            # pre-compute the features of the templates
            kp, des = self.kp_detector.detectAndCompute(template.image, None)
            self.all_template_info.append(TemplateInfo(template, kp, des))

    def clear_nmatch(self):
        for template_info in self.all_template_info:
            template_info.clear()
    
    def image_matching(self):
        if self.is_processing or self.image is None:
            return
        self.is_processing = True
        # compute the features of the current image
        self.kp, self.des = self.kp_detector.detectAndCompute(self.image, None)

        # get the match boxes of the current image
        current_template_info = self.all_template_info[self.likely_template_idx]
        final_results = self.match_with_template(current_template_info)

        max_n_matches = final_results['adjusted_n_matches']
        max_i = self.likely_template_idx
        # get the match boxes of the current template sliding window no matter what
        # template with the (matches + barrier) more matches than the current template
        #  will be selected (proactive switching)
        # or if the current template is not matched, match with first one that is matched
        for i in self.get_bi_indices(explore=False):
            results = self.match_with_template(self.all_template_info[i])
            i_adjusted_n_matches = results['adjusted_n_matches']
            i_boxes = results['boxes']
            if not (len(i_boxes[0]) > 0): # not a valid candidate for switching
                continue
            if (i_adjusted_n_matches > max_n_matches + WINDOW_BARRIER) \
                or (len(final_results['boxes'][0]) == 0):                 
                max_n_matches = i_adjusted_n_matches
                max_i = i
                final_results = results     

        # clear the n_matches if the template is changed
        # to avoid residual matches from the previous templates
        # that may make the template rapidly switch back by a lot
        if max_i != self.likely_template_idx and \
        (time.time() - self.last_change_template_time) > IMAGE_MATCHING_CHANGING_THRESHOLD:
            self.likely_template_idx = max_i
            self.last_change_template_time = time.time()
            self.clear_nmatch()

        # only if the all the templates in the sliding window are not matched, 
        # --> try to match other templates in two directions (reactive switching)
        if len(final_results['boxes'][0]) == 0:
            for i in self.get_bi_indices(explore=True):
                results = self.match_with_template(self.all_template_info[i])
                # several conditions to change the template
                # 1. there are enough matches
                # 2. the user is walking or spinning
                # 3. the time since the last change is larger than the threshold
                skip_penalty = abs(i - self.likely_template_idx)
                if len(results['boxes'][0]) > 0 \
                    and (results['adjusted_n_matches'] > max_n_matches + skip_penalty) \
                    and (self.walking or self.spinning) \
                    and (time.time() - self.last_change_template_time) > IMAGE_MATCHING_CHANGING_THRESHOLD * skip_penalty:
                    self.last_change_template_time = time.time()
                    # set the new likely template and clear the n_matches
                    self.likely_template_idx = i
                    final_results = results
                    self.clear_nmatch()
                    break
        
        # update the template information
        # draw the matching image
        self.n_matches = final_results['adjusted_n_matches']
        if valid_box(final_results['boxes'][0]):
            self.match_box = final_results['boxes'][0] 
            self.first_match = True
        else:
            self.match_box = []
        self.micro_adjust_dir.append(get_dir_from_boxes(final_results['boxes']))
        self.matching_img = self.draw_matches(self.n_matches,final_results['matches'], final_results['boxes'])
        
        self.is_processing = False

    def match_with_template(self,template):   
        matches = self._match(self.kp,template.kp, self.des,template.des)
        
        # adjust the number of matches using running average, to remove noise
        template.n_matches.append(len(matches))
        adjusted_n_matches =  template.n_matches.mean()

        boxes= self._get_boxes(self.kp,template.kp,matches,adjusted_n_matches)
        
        return dict(
            boxes=boxes,
            adjusted_n_matches=adjusted_n_matches,
            matches=matches,
        )

    def _match(self,kp1,kp2,des1, des2):
        if len(kp1) == 0 or len(kp2) == 0:
            return []
        matches = self.matcher.knnMatch(des1, des2, k=2)
        # eliminate bad matches
        new_matches = [m for m in matches if len(m) == 2 and m[0].distance < self.sim*m[1].distance]
        # new_matches = []
        # for m in matches:
        #     if len(m) == 2:
        #         kp1_this = kp1[m[0].queryIdx].pt
        #         kp2_this = kp2[m[0].trainIdx].pt
        #         dist = np.linalg.norm(np.array(kp1_this) - np.array(kp2_this))
        #         # 1. the distance between the matched keypoints is smaller than the threshold
        #         # 2. the distance in the image between those points is smaller than the threshold
        #         if m[0].distance < self.sim*m[1].distance and dist < self.kp_dist_threshold:
        #             new_matches.append(m)

        matches = new_matches
        return matches
    
    def _get_boxes(self, kp1, kp2, matches,adjusted_n_matches):

        # mathced keypoints in the current image
        target_kps = [tuple(int(j) for j in kp2[matches[i][0].trainIdx].pt) for i in range(len(matches))]
        target_kps = np.array(target_kps)
        match_box, image_box, template_box =[],[],[]
         
        # n_matches (adjusted from history) is larger than the threshold, the template is matched
        if adjusted_n_matches >=self.match_threshold:            
            # give it a box with four image corners 
            # if the true n_matches is smaller than the threshold (no box)
            if len(target_kps) < 4:
                image_box = (0,0, IMAGE_MATCHING_WIDTH, IMAGE_MATCHING_HEIGHT)
                template_box = image_box
                match_box = [0.5,0.5,0.5,0.5]
            else:
                # get the four corners of the matched keypoints from the template to form a box
                tr = target_kps.transpose(1,0)
                left_most_ind, top_most_ind  = np.argmin(tr, axis=1)
                right_most_ind, bottom_most_ind  = np.argmax(tr, axis=1)
                target_pts_ind  = [
                    left_most_ind, top_most_ind, 
                    right_most_ind, bottom_most_ind
                ]
                template_box = cv2.boundingRect(target_kps[target_pts_ind])

                # get the respective matched box in the current image
                query_pts = [tuple(int(j) for j in kp1[matches[i][0].queryIdx].pt) for i in target_pts_ind]
                image_box = cv2.boundingRect(np.array(query_pts))

                # in yolo box format, where the len are proportions, instead of absolute values
                # box is a rectangle (center_x, center_y, width, height)
                match_box = [
                    (image_box[0]+image_box[2]//2)/HEIGHT, 
                    (image_box[1]+ image_box[3]//2)/WIDTH, 
                    image_box[2]/HEIGHT, 
                    image_box[3]/WIDTH
                ]

        # get the box from the objects detected by yolo if image matching fails
        if len(match_box) == 0:
            match_box = self.get_box_from_objects()
        
        return match_box, image_box, template_box

    def get_bi_indices(self,explore=False,old = False): 
        '''
        get the template indices in two directions
        [0, 1, 2(current), 3, 4, 5, 6, 7, 8]
        the indices would be 1,3,4,etc until the radius is reached
        '''
        if not explore:
            if self.likely_template_idx == 0:
                return [1]
            elif self.likely_template_idx == len(self.all_template_info)-1:
                return [len(self.all_template_info)-2]
            else:
                lower =  self.likely_template_idx - 1
                upper =  self.likely_template_idx + 1
                return [lower, upper]

        whole_range =[]
        upper_done = False
        lower_done = False
        start_range = 1 if old else 2

        # if not matching for a long time, increase the searching radius
        add_to_radius = 0
        last_time = time.time()- self.last_change_template_time
        if last_time > IMAGE_MATCHING_EXPLORATION_THRESHOLD:
            add_to_radius += int((last_time // IMAGE_MATCHING_EXPLORATION_THRESHOLD)*2)

        for i in range(start_range, self.radius + 1 + add_to_radius):
            upper = min(len(self.all_template_info)-1, self.likely_template_idx + i)
            if not upper_done:
                whole_range.append(upper)   
            if upper == len(self.all_template_info)-1:
                upper_done = True

            lower = max(0, self.likely_template_idx - i)
            if not lower_done:
                whole_range.append(lower)
            # limit backtracking to 2 for efficiency
            if i ==2:
                lower_done = True

        return whole_range
    
    def get_box_from_objects(self):
        for box in self.detected_objects:
            x1, y1, x2, y2 = box['left'], box['top'], box['right'], box['bottom']
            x_c,y_c,w,h = xyxy2xywh_c(x1, y1, x2, y2)
            if box['id'] == self.all_template_info[self.likely_template_idx].template.cat:
                return [x_c,y_c,w,h]
        return []            

    def draw_matches(self,a_n_matches ,matches,boxes):
        likely_template_info = self.all_template_info[self.likely_template_idx]
        img1 = self.image
        cv2.putText(img1, str(a_n_matches), (10, 50), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 0, 255), 2)
        img2 = likely_template_info.template.image.copy()
        if len(boxes[0])!=0:
            boxes1a,boxes1b = (boxes[1][0], boxes[1][1]),  (boxes[1][2], boxes[1][3])
            boxes2a,boxes2b = (boxes[2][0], boxes[2][1]),  (boxes[2][2], boxes[2][3])
            cv2.rectangle(img1, boxes1a,boxes1b, (0, 0, 255), 2)
            cv2.rectangle(img2, boxes2a,boxes2b, (0, 0, 255), 2)
        img3 = cv2.drawMatchesKnn(
            img1, self.kp,
            img2, likely_template_info.kp,
            matches,None,**draw_params)
        return img3    
    