import numpy as np
import cv2
from os.path import join
import torch
import time
from algorithms.constants import HEIGHT, WIDTH
from algorithms.config import Config

class KeyPoint:
    def __init__(self, root, config_loader:Config, interval=0):
        self.enable=config_loader.image_path_following.enable
        self.engine_file = join(root, config_loader.image_path_following.model_file)   

        self.image=None

        self.kps = []
        self.desc= np.array([])

    def detectAndCompute(self,image, dummy=None):
        if not self.enable:
            return self.kps, self.desc
        valid = self.preprocess(image)
        self.detect()
        return self.kps, self.desc
        
    def preprocess(self,image,size=(WIDTH, HEIGHT)):

        resized_image = cv2.resize(image, size, cv2.INTER_LINEAR).astype(np.float32)

        # convert color
        if len(resized_image.shape) != 2:
            resized_image = cv2.cvtColor(resized_image, cv2.COLOR_BGR2GRAY)
        self.image= resized_image[np.newaxis, np.newaxis,:, :]/255.0
        
        return True
        
    def detect(self):
        if not self.enable:
            return
        self.infer()
    
    def destroy(self):
        pass


    def extract(self,semi , coarse_desc,cell=8 , border_remove=4 , nms_dist=4 , conf_thresh=0.015, nn_thresh=0.7):
        semi = semi.squeeze()
        # --- Process points.
        dense = np.exp(semi) # Softmax.
        dense = dense / (np.sum(dense, axis=0)+.00001) # Should sum to 1.
        # Remove dustbin.
        nodust = dense[:-1, :, :]
        # Reshape to get full resolution heatmap.
        Hc = int(HEIGHT / cell)
        Wc = int(WIDTH / cell)
        nodust = nodust.transpose(1, 2, 0)
        heatmap = np.reshape(nodust, [Hc, Wc, cell, cell])
        heatmap = np.transpose(heatmap, [0, 2, 1, 3])
        heatmap = np.reshape(heatmap, [Hc*cell, Wc*cell])
        xs, ys = np.where(heatmap >= conf_thresh) # Confidence threshold.
        if len(xs) == 0:
            return np.zeros((3, 0)), None, None
        pts = np.zeros((3, len(xs))) # Populate point data sized 3xN.
        pts[0, :] = ys
        pts[1, :] = xs
        pts[2, :] = heatmap[xs, ys]
        pts, _ = self.nms_fast(pts, HEIGHT, WIDTH, dist_thresh=nms_dist) # Apply NMS.
        inds = np.argsort(pts[2,:])
        pts = pts[:,inds[::-1]] # Sort by confidence.
        # Remove points along border.
        bord = border_remove
        toremoveW = np.logical_or(pts[0, :] < bord, pts[0, :] >= (WIDTH-bord))
        toremoveH = np.logical_or(pts[1, :] < bord, pts[1, :] >= (HEIGHT-bord))
        toremove = np.logical_or(toremoveW, toremoveH)
        pts = pts[:, ~toremove]

        # --- Process descriptor.
        D = coarse_desc.shape[1]
        if pts.shape[1] == 0:
            desc = np.zeros((D, 0))
        else:
            # pts.shape = (3, N), desc.shape = (1, 256, H/8, W/8)
            # Interpolate into descriptor map using 2D point locations.
            samp_pts = torch.from_numpy(pts[:2, :].copy())
            coarse_desc = torch.from_numpy(coarse_desc.copy())
            samp_pts[0, :] = (samp_pts[0, :] / (float(WIDTH)/2.)) - 1.
            samp_pts[1, :] = (samp_pts[1, :] / (float(HEIGHT)/2.)) - 1.
            samp_pts = samp_pts.transpose(0, 1).contiguous()
            samp_pts = samp_pts.view(1, 1, -1, 2)
            samp_pts = samp_pts.float()
            desc = torch.nn.functional.grid_sample(coarse_desc, samp_pts, align_corners=False)
            desc = desc.data.cpu().numpy().reshape(D, -1)
            desc /= np.linalg.norm(desc, axis=0)[np.newaxis, :]
            # desc.shape = (256, N)

        pts = pts.transpose(1,0)
        desc = desc.transpose(1,0)

        self.kps=self.convert_keypoints(pts)
        self.desc= self.convert_descriptors(desc)

    @staticmethod
    def convert_keypoints(keypoints):
        # turn the (x, y) coordinates of the keypoints into a cv2 KeyPoint
        return [cv2.KeyPoint(int(x), int(y), 1) for x, y,_ in keypoints]

    @staticmethod
    def convert_descriptors(descriptors):
        # descriptor range = [-norm, norm] -> [0, 255]
        # Scale the descriptors to the range [0, 255]
        v_min = np.min(descriptors)
        v_max = np.max(descriptors)
        descriptors = (descriptors - v_min) / (v_max - v_min) * 255
        # Convert the descriptors to 8-bit integers
        descriptors = descriptors.astype(np.uint8)
        return descriptors
    @staticmethod
    def nms_fast( in_corners, H, W, dist_thresh):
        """
        Run a faster approximate Non-Max-Suppression on numpy corners shaped:
        3xN [x_i,y_i,conf_i]^T
    
        Algo summary: Create a grid sized HxW. Assign each corner location a 1, rest
        are zeros. Iterate through all the 1's and convert them either to -1 or 0.
        Suppress points by setting nearby values to 0.
    
        Grid Value Legend:
        -1 : Kept.
        0 : Empty or suppressed.
        1 : To be processed (converted to either kept or supressed).
    
        NOTE: The NMS first rounds points to integers, so NMS distance might not
        be exactly dist_thresh. It also assumes points are within image boundaries.
    
        Inputs
        in_corners - 3xN numpy array with corners [x_i, y_i, confidence_i]^T.
        H - Image height.
        W - Image width.
        dist_thresh - Distance to suppress, measured as an infinty norm distance.
        Returns
        nmsed_corners - 3xN numpy matrix with surviving corners.
        nmsed_inds - N length numpy vector with surviving corner indices.
        """
        grid = np.zeros((H, W)).astype(int) # Track NMS data.
        inds = np.zeros((H, W)).astype(int) # Store indices of points.
        # Sort by confidence and round to nearest int.
        inds1 = np.argsort(-in_corners[2,:])
        corners = in_corners[:,inds1]
        rcorners = corners[:2,:].round().astype(int) # Rounded corners.
        # Check for edge case of 0 or 1 corners.
        if rcorners.shape[1] == 0:
            return np.zeros((3,0)).astype(int), np.zeros(0).astype(int)
        if rcorners.shape[1] == 1:
            out = np.vstack((rcorners, in_corners[2])).reshape(3,1)
            return out, np.zeros((1)).astype(int)
        # Initialize the grid.
        for i, rc in enumerate(rcorners.T):
            grid[rcorners[1,i], rcorners[0,i]] = 1
            inds[rcorners[1,i], rcorners[0,i]] = i
        # Pad the border of the grid, so that we can NMS points near the border.
        pad = dist_thresh
        grid = np.pad(grid, ((pad,pad), (pad,pad)), mode='constant')
        # Iterate through points, highest to lowest conf, suppress neighborhood.
        count = 0
        for i, rc in enumerate(rcorners.T):
            # Account for top and left padding.
            pt = (rc[0]+pad, rc[1]+pad)
            if grid[pt[1], pt[0]] == 1: # If not yet suppressed.
                grid[pt[1]-pad:pt[1]+pad+1, pt[0]-pad:pt[0]+pad+1] = 0
                grid[pt[1], pt[0]] = -1
                count += 1
        # Get all surviving -1's and return sorted array of remaining corners.
        keepy, keepx = np.where(grid==-1)
        keepy, keepx = keepy - pad, keepx - pad
        inds_keep = inds[keepy, keepx]
        out = corners[:, inds_keep]
        values = out[-1, :]
        inds2 = np.argsort(-values)
        out = out[:, inds2]
        out_inds = inds1[inds_keep[inds2]]
        return out, out_inds
    