
import cv2
import os
import requests
import json
from algorithms.constants import IDLE_THRESHOLD,MAIN_SERVER,MAIN_SERVER_PORT
from algorithms.general.internet_checker import ping_server
from algorithms.general.utils import PeriodicThread,get_bool_from_txt,write_txt
from algorithms.general.auth import AuthProvider
from algorithms.config import Config

from .utils import *
from .bi_image_matcher import BidirectionalImageMatcher

class Trip:
    def __init__(self, trip_id, trip_name, keyframes):
        self.trip_id = trip_id
        self.trip_name = trip_name
        self.keyframes = keyframes
        self.trip_len = len(keyframes)

    def give_keyframe_desc(self, idx):
        if idx < len(self.keyframes) and idx >= 0:
            return (
                self.keyframes[idx]["is_decision_point"],
                self.keyframes[idx]["description"],
            )
        return False, ""

    # get the turning direction of the template in the context of the trip
    def get_turning(self, idx):
        if idx < len(self.keyframes) and idx >= 0:
            return self.keyframes[idx]["turning_direction"]
        return 0.0


class MatchPropeller:
    def __init__(
        self,
        root: str,
        config: Config,
        logger,
        auth_provider: AuthProvider = None,
    ):
        self.root = root
        self.config = config
        self.logger = logger

        self.auth_provider = auth_provider

        self.trip = None
        self.init_model()
        self.reconfigure(config)
        
    def init_model(self):
        if not self.config.image_path_following.use_superpoint:
            self.keypoints_detector = None
            return
        
        if self.config.engine_type.lower() == "onnx":
            from .keypoints.keypoints_onnx import KeyPointONNX as KeyPoint
        elif self.config.engine_type.lower() == "bpu":
            from .keypoints.keypoints_bpu import KeyPointBPU as KeyPoint
        self.keypoints_detector = KeyPoint(self.root, self.config)
        

    def reconfigure(self, config: Config):
        self.config = config
        self.matcher = BidirectionalImageMatcher(self.keypoints_detector)
        # self.start()

    def start(self):
        if not self.config.image_path_following.enable:
            return
        self.receive_all_template()
        # self.trip_reset_file = os.path.join(self.root, "global_states", "trip_reset")
        # self.trip_reload_file = os.path.join(self.root, "global_states", "trip_reload")
        # self.trip_forward_file = os.path.join(self.root, "global_states", "trip_forward")
        # self.trip_backward_file = os.path.join(self.root, "global_states", "trip_backward")
        # self.reset_operation_states()

        self.check_trip_operation_thread = PeriodicThread(1, self.check_trip_operation)
        self.image_matching_thread = PeriodicThread(0.05, self.matcher.image_matching)

        self.check_trip_operation_thread.start()
        self.image_matching_thread.start()
        
    def stop(self):
        if not self.config.image_path_following.enable:
            return
        self.check_trip_operation_thread.stop() if self.check_trip_operation_thread else None
        self.image_matching_thread.stop() if self.image_matching_thread else None
        
    def release(self):
        self.stop()
        if self.keypoints_detector is not None:
            self.keypoints_detector.destroy()

    # Update the user's coordinate and others
    def update(self, input_frame, motion_data, objects):
        if not self.config.image_path_following.enable:
            return
        self.matcher.set_image(input_frame)

        walking_value, spinning_value = motion_data
        self.matcher.set_state(
            walking_value > IDLE_THRESHOLD, spinning_value > IDLE_THRESHOLD
        )

        if objects is not None:
            self.matcher.detected_objects = objects

    def get_data(self):
        return (
            self.get_matcher_states(),
            self.get_template_info(),
        )

    def get_matcher_states(self):
        if not self.config.image_path_following.enable:
            return None
        align = True if self.matcher.match_box != [] else False
        template_idx = self.matcher.likely_template_idx
        finished_trip = (
            template_idx >= self.trip.trip_len - 1
            if template_idx is not None
            else False
        )
        return dict(
            align=align,
            first_match=self.matcher.first_match,
            match_box=self.matcher.match_box,
            template_idx=template_idx,
            finished_trip=finished_trip,
            matching_img=self.matcher.matching_img,
            n_matches=self.matcher.n_matches,
            micro_adjust_dir=self.matcher.micro_adjust_dir.mean(),
        )

    def get_template_info(self):
        if not self.config.image_path_following.enable:
            return None
        is_decision_point, desc = self.trip.give_keyframe_desc(
            self.matcher.likely_template_idx
        )
        turning_dir = self.trip.get_turning(self.matcher.likely_template_idx)
        return dict(
            is_decision_point=is_decision_point,
            keyframe_desc=desc,
            turning_dir=turning_dir,
        )

    def receive_all_template(self):
        if self.config.image_path_following.mock_route:
            fake_temp_dir = self.config.image_path_following.mock_route_folder
            if not os.path.exists(fake_temp_dir):
                self.logger.info("Mock Route Folder Not Found")
                return
            with open(os.path.join(fake_temp_dir, "config.json")) as f:
                config = json.load(f)
            # sorted using numerical order
            all_templates = []
            for i in sorted(
                [
                    j
                    for j in os.listdir(fake_temp_dir)
                    if j.split(".")[-1] == "jpg" or j.split(".")[-1] == "png"
                ],
                key=lambda x: int(x.split(".")[0]),
            ):
                image = cv2.imread(os.path.join(fake_temp_dir, i), 0)
                all_templates.append(
                    dict(
                        image_id=i.split("/")[-1].split(".")[0],
                        image=image,
                        category=None,
                        direction=None,
                    )
                )
            trip_id = "fake_trip"
            trip_name = config["name"]
            keyframes = config["keyframes"]
        else:
            assert ping_server(
                MAIN_SERVER, MAIN_SERVER_PORT, timeout=3
            ), "MAIN_SERVER Not Responsive, Please disable online matching"
            trip_id, trip_name, keyframes, all_templates = self.get_trip_config()
            if len(all_templates) == 0:
                self.logger.info("Matching Server Error, No Template Returned")
                return

        self.trip = Trip(trip_id, trip_name, keyframes)
        self.matcher.set_all_template(all_templates)

    def get_trip_config(self):
        header, device_info = self.auth_provider.get_header_and_device_info()        
        current_trip_id =device_info["current_trip"]
        
        
        cache_dir = os.path.join(self.root, ".cache")
        os.makedirs(cache_dir, exist_ok=True)

        cache_file = os.path.join(cache_dir, f"{current_trip_id}.json")
        downloaded = os.path.exists(cache_file)
        # get the config from the cache if the id is the same
        if downloaded:
            # get the config from the server (without the images)
            new_config = self.get_new_config(header, current_trip_id, with_images=False)

            # load the cached config
            with open(cache_file, "r") as f:
                old_config = json.load(f)

            # the config is different, get the new config with images from the server
            if len(old_config["keyframes"]) != len(new_config["keyframes"]):
                self.logger.info(
                    "Trip Config Length Changed, Downloading New Trip Config"
                )
                new_config = self.get_new_config(header, current_trip_id, with_images=True)
            else:
                self.logger.info("Using Cached Trip Config")
                # transfer the image from the old config to the new config
                for old_kf, new_kf in zip(
                    old_config["keyframes"], new_config["keyframes"]
                ):
                    new_kf["db_image"] = old_kf["db_image"]
        else:
            self.logger.info("Downloading New Trip Config.")
            # get the config from the server
            new_config = self.get_new_config(header, current_trip_id, with_images=True)

        trip_id = current_trip_id
        trip_name = new_config["name"]
        keyframes = new_config["keyframes"]

        # save the new config to the cache
        with open(cache_file, "w") as f:
            json.dump(new_config, f)

        # convert the base64 string to image
        images = [kf["db_image"] for kf in keyframes]
        for i in images:
            i["image"] = base642img(i["image_base64"])

        # remove the db_image from the keyframes to save space
        for kf in keyframes:
            del kf["db_image"]

        return trip_id, trip_name, keyframes, images

    @staticmethod
    def get_new_config(header, trip_id, with_images):
        try:
            res = requests.get(
                f"{MAIN_SERVER}/trip/list_trip",
                headers=header,
                params={"trip_id": trip_id, "with_images": with_images},
                timeout=5,
            )
            if res.status_code != 200:
                return {"error": "could not get config"}
            new_config = res.json()['trips'][str(trip_id)]
        except:
            return {"error": "could not get config"}
        return new_config
    # @ staticmethod
    # def reset_txt(path):
    #     write_txt(path,'0')


    def check_trip_operation(self):
        pass
    #     need_to_reset = get_bool_from_txt(self.trip_reset_file)
    #     need_to_reload = get_bool_from_txt(self.trip_reload_file)
    #     need_to_forward = get_bool_from_txt(self.trip_forward_file)
    #     need_to_backward = get_bool_from_txt(self.trip_backward_file)

    #     # reset matcher
    #     if not (need_to_reset or need_to_reload or need_to_forward or need_to_backward):
    #         return
    #     # wait for the matcher to finish processing
    #     while self.matcher.is_processing:
    #         pass

    #     self.matcher.is_processing = True
    #     if need_to_forward:
    #         self.matcher.forward()
    #         self.logger.info("Trip Called To Forward!")
    #     elif need_to_backward:
    #         self.matcher.backward()
    #         self.logger.info("Trip Called To Roll Back!")
    #     else:
    #         self.matcher.reset()
    #         self.logger.info("Trip Index Called To Reset!")
    #         # avoid calling the matcher when loading the templates
    #         if need_to_reload:
    #             self.logger.info("Trip Called To Reload!")
    #             self.receive_all_template()

    #     self.matcher.is_processing = False
    #     self.reset_operation_states()

    # def reset_operation_states(self):
    #     # reset all states
    #     self.reset_txt(self.trip_reset_file)
    #     self.reset_txt(self.trip_reload_file)
    #     self.reset_txt(self.trip_forward_file)
    #     self.reset_txt(self.trip_backward_file)
