from rclpy.node import Node

from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition
from lifecycle_msgs.msg import TransitionEvent
from lifecycle_msgs.msg import TransitionDescription

from lifecycle_msgs.srv import ChangeState
from lifecycle_msgs.srv import GetAvailableStates
from lifecycle_msgs.srv import GetAvailableTransitions
from lifecycle_msgs.srv import GetState


def get_label( msg_type, id):
    for key, value in vars(msg_type).items():
        if(value == id):
            return key
    return None

def state_cls(state):
    return State(
        id=state,
        label=get_label(State, state)
    )
def transition_cls(transition):
    return Transition(
        id=transition,
        label=get_label(Transition, transition),
     
    )
def transition_des_cls(transition):
    return TransitionDescription(
        transition=transition_cls(transition),
        start_state=state_cls(State.PRIMARY_STATE_UNKNOWN),
        goal_state=state_cls(State.PRIMARY_STATE_UNKNOWN)
    )

class LifecycleNode(Node):    
    def __init__(self, node_name:str):
        super().__init__(node_name)
        self.state = State.PRIMARY_STATE_UNKNOWN

        self.available_transitions = []
        self.available_states = [
            state_cls(State.PRIMARY_STATE_UNKNOWN),
            state_cls(State.PRIMARY_STATE_UNCONFIGURED),
            state_cls(State.PRIMARY_STATE_INACTIVE),
            state_cls(State.PRIMARY_STATE_ACTIVE),
            state_cls(State.PRIMARY_STATE_FINALIZED),
            state_cls(State.TRANSITION_STATE_CONFIGURING),
            state_cls(State.TRANSITION_STATE_CLEANINGUP),
            state_cls(State.TRANSITION_STATE_SHUTTINGDOWN),
            state_cls(State.TRANSITION_STATE_ACTIVATING),
            state_cls(State.TRANSITION_STATE_DEACTIVATING),
            state_cls(State.TRANSITION_STATE_ERRORPROCESSING)
        ]

        self.srv_get_state = self.create_service(
                GetState, 
                node_name + '/get_state',
                self.get_state
            )

        self.srv_change_state = self.create_service(
                ChangeState,
                node_name + '/change_state',
                self.change_state
            )

        self.srv_get_available_states = self.create_service(
                GetAvailableStates, 
                node_name + '/get_available_states',
                self.get_available_states
            )

        self.srv_get_available_transitions = self.create_service(
                GetAvailableTransitions,
                node_name + '/get_available_transitions',
                self.get_available_transitions
            )

        self.pub_transition_event = self.create_publisher(
                TransitionEvent, 
                node_name + '/transition_event',
                1
            )
        
        self.create()
        
    def get_state(self, request: GetState.Request, response: GetState.Response):
        response.current_state = state_cls(self.state)
        return response 
    

    def change_state(self, request : ChangeState.Request, response : ChangeState.Response):
        
        if(request.transition.id == Transition.TRANSITION_CREATE):
            response.success = (self.create() == Transition.TRANSITION_CALLBACK_SUCCESS)

        elif(request.transition.id == Transition.TRANSITION_CONFIGURE):
            response.success = (self.configure() == Transition.TRANSITION_CALLBACK_SUCCESS)

        elif(request.transition.id == Transition.TRANSITION_CLEANUP):
            response.success = (self.cleanup() == Transition.TRANSITION_CALLBACK_SUCCESS)

        elif(request.transition.id == Transition.TRANSITION_ACTIVATE):
            response.success = (self.activate() == Transition.TRANSITION_CALLBACK_SUCCESS)

        elif(request.transition.id == Transition.TRANSITION_DEACTIVATE):
            response.success = (self.deactivate() == Transition.TRANSITION_CALLBACK_SUCCESS)

        elif(request.transition.id == Transition.TRANSITION_UNCONFIGURED_SHUTDOWN
              or request.transition.id == Transition.TRANSITION_INACTIVE_SHUTDOWN
              or request.transition.id == Transition.TRANSITION_ACTIVE_SHUTDOWN):
            response.success = (self.shutdown() == Transition.TRANSITION_CALLBACK_SUCCESS)

        elif(request.transition.id==Transition.TRANSITION_DESTROY):
            response.success = self.destroy()
        else:
            response.success = False

        return response


    def get_available_states(self, request : GetAvailableStates.Request, response : GetAvailableStates.Response):
        response.available_states = self.available_states
        return response


    def get_available_transitions(self, request : GetAvailableTransitions.Request, response : GetAvailableTransitions.Response):
        response.available_transitions = self.available_transitions
        return response


    def create(self):
        if(self.state == State.PRIMARY_STATE_UNKNOWN):
            self.state = State.PRIMARY_STATE_UNCONFIGURED
            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(Transition.TRANSITION_CREATE),
                    start_state = state_cls(State.PRIMARY_STATE_UNKNOWN),
                    goal_state = state_cls(self.state )
                )
            )
            self.available_transitions = [
                transition_des_cls(Transition.TRANSITION_CONFIGURE),
                transition_des_cls(Transition.TRANSITION_UNCONFIGURED_SHUTDOWN)
            ]

            return Transition.TRANSITION_CALLBACK_SUCCESS 
        else:
            return Transition.TRANSITION_CALLBACK_FAILURE


    def configure(self):
        if(self.state == State.PRIMARY_STATE_UNCONFIGURED):

            self.state = State.TRANSITION_STATE_CONFIGURING

            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(Transition.TRANSITION_CONFIGURE),
                    start_state = state_cls(State.PRIMARY_STATE_UNCONFIGURED),
                    goal_state = state_cls(self.state)
                )
            )

            task_config = self.executor.create_task(self.on_configure)

            self.executor.spin_until_future_complete(task_config)

            result_transition = None
            if(task_config.result() == Transition.TRANSITION_CALLBACK_SUCCESS):
                self.state = State.PRIMARY_STATE_INACTIVE
                result_transition = Transition.TRANSITION_ON_CONFIGURE_SUCCESS
                self.available_transitions = [
                    transition_des_cls(Transition.TRANSITION_ACTIVATE),
                    transition_des_cls(Transition.TRANSITION_CLEANUP),
                    transition_des_cls(Transition.TRANSITION_INACTIVE_SHUTDOWN),
                ]

            elif(task_config.result() == Transition.TRANSITION_CALLBACK_FAILURE):
                self.state = State.PRIMARY_STATE_UNCONFIGURED
                result_transition = Transition.TRANSITION_ON_CONFIGURE_FAILURE
                self.available_transitions = [
                    transition_des_cls(Transition.TRANSITION_CONFIGURE),
                    transition_des_cls(Transition.TRANSITION_UNCONFIGURED_SHUTDOWN)
                ]

            else:
                self.state = State.TRANSITION_STATE_ERRORPROCESSING
                result_transition = Transition.TRANSITION_ON_CONFIGURE_ERROR

            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(result_transition),
                    start_state = state_cls(State.PRIMARY_STATE_UNCONFIGURED),
                    goal_state = state_cls(self.state)
                )
            )       

            return task_config.result()

        else:
            return Transition.TRANSITION_CALLBACK_FAILURE


    def cleanup(self):
        if(self.state == State.PRIMARY_STATE_INACTIVE):

            self.state = State.TRANSITION_STATE_CLEANINGUP

            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(Transition.TRANSITION_CLEANUP),
                    start_state = state_cls(State.PRIMARY_STATE_INACTIVE),
                    goal_state = state_cls(self.state)
                )
            )       

            task_cleanup = self.executor.create_task(self.on_cleanup)

            self.executor.spin_until_future_complete(task_cleanup)

            result_transition = None
            if(task_cleanup.result() == Transition.TRANSITION_CALLBACK_SUCCESS):
                self.state = State.PRIMARY_STATE_UNCONFIGURED
                result_transition = Transition.TRANSITION_ON_CLEANUP_SUCCESS
                self.available_transitions = [
                    transition_des_cls(Transition.TRANSITION_CONFIGURE),
                    transition_des_cls(Transition.TRANSITION_UNCONFIGURED_SHUTDOWN)
                ]

            else:
                self.state = State.TRANSITION_STATE_ERRORPROCESSING
                result_transition = Transition.TRANSITION_ON_CLEANUP_ERROR
            
            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(result_transition),
                    start_state = state_cls(State.TRANSITION_STATE_CLEANINGUP),
                    goal_state =  state_cls(self.state)
                )
            )       

            return task_cleanup.result()

        else:
            return Transition.TRANSITION_CALLBACK_FAILURE


    def activate(self):
        if(self.state == State.PRIMARY_STATE_INACTIVE):

            self.state = State.TRANSITION_STATE_ACTIVATING

            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition =transition_cls(Transition.TRANSITION_ACTIVATE),
                    start_state =state_cls(State.PRIMARY_STATE_INACTIVE),
                    goal_state =  state_cls(self.state)
                )
            )       

            task_activate = self.executor.create_task(self.on_activate)

            self.executor.spin_until_future_complete(task_activate)

            result_transition = None
            if(task_activate.result() == Transition.TRANSITION_CALLBACK_SUCCESS):
                self.state = State.PRIMARY_STATE_ACTIVE
                result_transition = Transition.TRANSITION_ON_ACTIVATE_SUCCESS
                self.available_transitions = [
                    transition_des_cls(Transition.TRANSITION_DEACTIVATE),
                    transition_des_cls(Transition.TRANSITION_ACTIVE_SHUTDOWN),
                ]

            elif(task_activate.result() == Transition.TRANSITION_CALLBACK_FAILURE):
                self.state = State.PRIMARY_STATE_INACTIVE
                result_transition = Transition.TRANSITION_ON_ACTIVATE_FAILURE
                self.available_transitions = [
                    transition_des_cls(Transition.TRANSITION_ACTIVATE),
                    transition_des_cls(Transition.TRANSITION_CLEANUP),
                    transition_des_cls(Transition.TRANSITION_INACTIVE_SHUTDOWN),
                ]

            else:
                self.state = State.TRANSITION_STATE_ERRORPROCESSING
                result_transition = Transition.TRANSITION_ON_CONFIGURE_ERROR
            
            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition =transition_cls(result_transition),
                    start_state = state_cls( State.TRANSITION_STATE_ACTIVATING),
                    goal_state = state_cls(self.state)
                )
            )       

            return task_activate.result()

        else:
            return Transition.TRANSITION_CALLBACK_FAILURE


    def deactivate(self):
        if(self.state == State.PRIMARY_STATE_ACTIVE):

            self.state = State.TRANSITION_STATE_DEACTIVATING

            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(Transition.TRANSITION_DEACTIVATE),
                    start_state = state_cls(State.PRIMARY_STATE_ACTIVE),
                    goal_state = state_cls(self.state)
                )
            )       

            task_deactivate = self.executor.create_task(self.on_deactivate)

            self.executor.spin_until_future_complete(task_deactivate)

            result_transition = None
            if(task_deactivate.result() == Transition.TRANSITION_CALLBACK_SUCCESS):
                self.state = State.PRIMARY_STATE_INACTIVE
                result_transition = Transition.TRANSITION_ON_DEACTIVATE_SUCCESS
                self.available_transitions = [
                    transition_des_cls(Transition.TRANSITION_ACTIVATE),
                    transition_des_cls(Transition.TRANSITION_CLEANUP),
                    transition_des_cls(Transition.TRANSITION_INACTIVE_SHUTDOWN),
                ]

            else:
                self.state = State.TRANSITION_STATE_ERRORPROCESSING
                result_transition = Transition.TRANSITION_ON_DEACTIVATE_ERROR
            
            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(result_transition),
                    start_state = state_cls(State.TRANSITION_STATE_DEACTIVATING),
                    goal_state = state_cls(self.state)
                )
            )       

            return task_deactivate.result()

        else:
            return Transition.TRANSITION_CALLBACK_FAILURE


    def shutdown(self):
        if (self.state == State.PRIMARY_STATE_UNCONFIGURED):
            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(Transition.TRANSITION_UNCONFIGURED_SHUTDOWN),
                    start_state = state_cls(State.PRIMARY_STATE_UNCONFIGURED),
                    goal_state = state_cls(State.TRANSITION_STATE_SHUTTINGDOWN)
                )
            )       

        elif (self.state == State.PRIMARY_STATE_INACTIVE):
            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition =  transition_cls(Transition.TRANSITION_INACTIVE_SHUTDOWN),
                    start_state = state_cls(State.PRIMARY_STATE_INACTIVE),
                    goal_state = state_cls(State.TRANSITION_STATE_SHUTTINGDOWN) 
                )
            )       

        elif (self.state == State.PRIMARY_STATE_ACTIVE):
            self.pub_transition_event.publish(
                TransitionEvent(
                    timestamp = self.get_clock().now().nanoseconds,
                    transition = transition_cls(Transition.TRANSITION_ACTIVE_SHUTDOWN),
                    start_state = state_cls(State.PRIMARY_STATE_ACTIVE),
                    goal_state = state_cls(State.TRANSITION_STATE_SHUTTINGDOWN)
                )
            )       

        else:
            return Transition.TRANSITION_CALLBACK_FAILURE

        self.state = State.TRANSITION_STATE_SHUTTINGDOWN

        task_shutdown = self.executor.create_task(self.on_shutdown)

        self.executor.spin_until_future_complete(task_shutdown)

        result_transition = None
        if(task_shutdown.result() == Transition.TRANSITION_CALLBACK_SUCCESS):
            self.state = State.PRIMARY_STATE_FINALIZED
            result_transition = Transition.TRANSITION_ON_SHUTDOWN_SUCCESS
            self.available_transitions=[transition_des_cls(Transition.TRANSITION_DESTROY)]
        else:
            self.state = State.TRANSITION_STATE_ERRORPROCESSING
            result_transition = Transition.TRANSITION_ON_SHUTDOWN_ERROR
        
        self.pub_transition_event.publish(
            TransitionEvent(
                timestamp = self.get_clock().now().nanoseconds,
                transition = transition_cls(result_transition),
                start_state = state_cls(State.TRANSITION_STATE_SHUTTINGDOWN),
                goal_state = state_cls(self.state)
            )
        )       

        return task_shutdown.result()


    def destroy(self):
        if(self.state == State.PRIMARY_STATE_FINALIZED):
            self.logger.info("Destroying node")
            self.destroy_node()
            # exit the rclpy.spin as well
            self.executor.shutdown()
        else:
            return Transition.TRANSITION_CALLBACK_FAILURE


    def on_configure(self):
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_cleanup(self):
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_activate(self):
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_deactivate(self):
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_error(self):
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_shutdown(self):
        return Transition.TRANSITION_CALLBACK_SUCCESS