import threading
from os.path import dirname, abspath, join
import numpy as np
from typing import List
import gc

from ament_index_python.packages import get_package_share_directory
import rclpy
from rclpy.time import Time
from rclpy.node import Node, SetParametersResult, Parameter
from std_msgs.msg import Float64MultiArray, String
from sensor_msgs.msg import Image, CompressedImage
from rclpy.qos import QoSPresetProfiles
from cv_bridge import CvBridge
from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition
from interfaces.msg import (
    ObstacleAvoidanceData,
    ObjectTrackingData,
    ImageMatchingData,
    MapPathFollowingData,
    CloudData,
)

from algorithms.lifecycle import LifecycleNode
from algorithms.builder import (
    get_algorithms,
    get_io_and_utils,
    get_ms_engines,
)

from algorithms.constants import (
    SHORT_TERM_IDLE_TIMEOUT,
)
from algorithms.config import (
    sample_config,
    find_param_str_in_config,
    set_param_str_in_config,
)

ROOT_PATH = join(
    get_package_share_directory("algorithms"),
)
# DATA_QOS = QoSPresetProfiles.SENSOR_DATA.value
DATA_QOS = 1
RELIABLE_QOS = QoSPresetProfiles.SYSTEM_DEFAULT.value


class Algorithms(LifecycleNode):
    def __init__(self):
        super().__init__("algorithms_node")
        self.declare_parameters(
            namespace="",
            parameters=[
                ("namespace", None),
                ("imu_namespace", None),
                ("camera0_namespace", None),
                ("camera1_namespace", None),
                ("rate", None),
                ("engine_type", None),
                ("obstacle_avoidance.enable", None),
                ("obstacle_avoidance.model_file", None),
                ("obstacle_avoidance.black_and_white", None),
                ("obstacle_avoidance.alpha", None),
                ("obstacle_avoidance.use_obstacle_detect", None),
                ("obstacle_avoidance.use_apf_route", None),
                ("object_tracking.enable", None),
                ("object_tracking.model_file", None),
                ("object_tracking.use_tracking", None),
                ("image_path_following.enable", None),
                ("image_path_following.mock_route", None),
                ("image_path_following.mock_route_folder", None),
                ("image_path_following.use_superpoint", None),
                ("image_path_following.model_file", None),
                ("map_path_following.enable", None),
                ("map_path_following.mock_route", None),
                ("map_path_following.mock_route_folder", None),
                ("cloud.image_captioning", None),
                ("cloud.image_tagging", None),
                ("cloud.scene_text_spotting", None),
                ("visualize.enable", None),
                ("visualize.debug", None),
                ("text_to_speech.enable", None),
                ("text_to_speech.interval", None),
            ],
        )
        self.logger = self.get_logger()
        self.config = None
        self.lock = threading.Lock()
        self.add_on_set_parameters_callback(self.config_update)

    ############################## Lifecycle ##############################
    def on_configure(self):
        self.logger.info("on_configure() is called")
        if self.config is None:
            self.config = sample_config()
            for param in self.get_parameters_by_prefix("").keys():
                set_param_str_in_config(
                    self.config, param, self.get_parameter(param).value
                )

        self.is_processing = False
        self.bridge = CvBridge()
        self.frame_raw0 = None
        self.frame_raw1 = None
        self.last_frame0_update = self.get_clock().now()
        self.last_frame1_update = self.get_clock().now()
        
        self.imu_data = None
        self.ori_data = None
        self.gps_data = None

        self.object_detected = None
        self.loop_vis = []
        self.is_decision_point = False
        
        try:
            with self.lock:
                self.build_childern()
        except Exception as e:
            self.logger.error(f"Init failed: {e}")
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_cleanup(self):
        self.logger.info("on_cleanup() is called")
        self.cleanup_childern()
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_activate(self):
        self.logger.info("on_activate() is called")
        self.childern_complete = False
        
        # =================== Publishers ===================
        self.ann0_pub = self.create_publisher(
            Image, f"{self.config.namespace}/annotation0",DATA_QOS
        )
        self.ann1_pub = self.create_publisher(
            Image, f"{self.config.namespace}/annotation1", DATA_QOS
        )
        self.obs_data_pub = self.create_publisher(
            ObstacleAvoidanceData, f"{self.config.namespace}/obs_data", DATA_QOS
        )
        self.obj_data_pub = self.create_publisher(
            ObjectTrackingData, f"{self.config.namespace}/obj_data", DATA_QOS
        )
        self.match_data_pub = self.create_publisher(
            ImageMatchingData, f"{self.config.namespace}/match_data", DATA_QOS
        )
        self.map_data_pub = self.create_publisher(
            MapPathFollowingData, f"{self.config.namespace}/map_data", DATA_QOS
        )
        self.cloud_data_pub = self.create_publisher(
            CloudData, f"{self.config.namespace}/cloud_data", DATA_QOS
        )
        self.tts_text = self.create_publisher(String, "/tts_text", 10)

        # =================== Subscribers ===================
        self.cam0_sub = self.create_subscription(
            Image,
            f"{self.config.camera0_namespace}/image_raw",
            self.frame0_update,
            DATA_QOS,
        )
        self.cam1_sub = self.create_subscription(
            Image,
            f"{self.config.camera1_namespace}/image_raw",
            self.frame1_update,
            DATA_QOS,
        )
        self.imu_sub = self.create_subscription(
            Float64MultiArray,
            f"{self.config.imu_namespace}/imu",
            self.imu_update,
            DATA_QOS,
        )
        self.ori_sub = self.create_subscription(
            Float64MultiArray,
            f"{self.config.imu_namespace}/orientation",
            self.ori_udpate,
            DATA_QOS,
        )
        self.gps_sub = self.create_subscription(
            Float64MultiArray,
            f"{self.config.imu_namespace}/gps",
            self.gps_update,
            DATA_QOS,
        )

        # =================== Timer ===================
        rate = self.config.rate
        self.timer = self.create_timer(1 / rate, self.process_frame)
        
        self.update_childern_params()
        self.start_childern()
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_deactivate(self):
        self.logger.info("on_deactivate() is called")
        # wait for the last timer callback to finish
        self.timer.cancel()
        while self.is_processing:
            pass
        self.stop_childern()
        self.timer = None

        self.ann0_pub = None
        self.ann1_pub = None
        self.obs_data_pub = None
        self.obj_data_pub = None
        self.match_data_pub = None
        self.map_data_pub = None
        self.cloud_data_pub = None
        self.tts_text = None

        self.cam0_sub = None
        self.cam1_sub = None
        self.imu_sub = None
        self.ori_sub = None
        self.gps_sub = None

        gc.collect()
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def destroy_node(self) -> bool:
        # TODO: the terminate function
        self.stop_childern()
        self.cleanup_childern()
        return super().destroy_node()

    ############################## Config Update ##############################
    def config_update(self, params: List[Parameter]):
        self.logger.info(f"Update Algorithms: {[str(p.name) for p in params]}")
        for param in params:
            config_param_value = find_param_str_in_config(self.config, param.name)
            if config_param_value is not None:
                if param.value == config_param_value:
                    continue
                set_param_str_in_config(self.config, param.name, param.value)
            else:
                self.logger.warn(f"Unknown parameter: {param.name}")
        return SetParametersResult(successful=True)

    ############################## Childern Lifecycles ##############################
    def update_childern_params(self):
        if not self.check_childern_complete():
            self.logger.warn("Not all children are initialized")
            return
        self.obstacle_avoidance .reconfigure(self.config)
        self.object_tracker     .reconfigure(self.config)
        self.match_tracker      .reconfigure(self.config)
        self.path_tracker       .reconfigure(self.config)
        self.scene_text_spotter .reconfigure(self.config.cloud.scene_text_spotting)
        self.image_captioner    .reconfigure(self.config.cloud.image_captioning)
        self.image_tagger       .reconfigure(self.config.cloud.image_tagging)
        self.visualize          .reconfigure(self.config)
        self.text_to_speech     .reconfigure(self.config)
        self.logger.info("Main Algos Reconfigured")

    def build_childern(self):
        self.cleanup_childern()
        # general io and utils
        (
            self.visualize,
            self.text_to_speech,
            self.garbage_collector,
            self.auto_idle,
            self.internet_checker,
            self.auth_provider,
            self.fps_er,
        ) = get_io_and_utils(self.config, self.logger)

        (
            self.obstacle_avoidance,
            self.object_tracker,
            self.match_tracker,
            self.path_tracker,
        ) = get_algorithms(
            ROOT_PATH,
            self.config,
            self.logger,
            self.auth_provider,
        )
        # ms demo engines
        (
            self.scene_text_spotter,
            self.image_captioner,
            self.image_tagger,
            self.cloud_preprocessor,
        ) = get_ms_engines(self.config, self.logger)
        if not self.check_childern_complete():
            self.logger.warn("Not all children are initialized")
        else:
            self.logger.info("Main Algos childern built")

    def cleanup_childern(self):
        try: 
            self.obstacle_avoidance .release() if hasattr(self, "obstacle_avoidance") else None
            self.object_tracker     .release() if hasattr(self, "object_tracker") else None
            self.match_tracker      .release() if hasattr(self, "match_tracker") else None
            self.path_tracker       .release() if hasattr(self, "path_tracker") else None
            self.cloud_preprocessor .release() if hasattr(self, "cloud_preprocessor") else None

            self.internet_checker   .release() if hasattr(self, "internet_checker") else None
            self.visualize          .release() if hasattr(self, "visualize") else None
            self.garbage_collector  .instant_collect() if hasattr(self, "garbage_collector") else None
        except Exception as e:
            raise e
        self.logger.info("Main Algos childern cleaned up")

    def start_childern(self):
        self.obstacle_avoidance .start() if hasattr(self, "obstacle_avoidance") else None
        self.object_tracker     .start() if hasattr(self, "object_tracker") else None
        self.match_tracker      .start() if hasattr(self, "match_tracker") else None
        self.path_tracker       .start() if hasattr(self, "path_tracker") else None
        self.cloud_preprocessor .start() if hasattr(self, "cloud_preprocessor") else None

        self.internet_checker   .start() if hasattr(self, "internet_checker") else None
        self.visualize          .start() if hasattr(self, "visualize") else None
        self.logger.info("Main Algos childern started")

    def stop_childern(self):
        self.obstacle_avoidance .stop() if hasattr(self, "obstacle_avoidance") else None
        self.object_tracker     .stop() if hasattr(self, "object_tracker") else None
        self.match_tracker      .stop() if hasattr(self, "match_tracker") else None
        self.path_tracker       .stop() if hasattr(self, "path_tracker") else None
        self.cloud_preprocessor .stop() if hasattr(self, "cloud_preprocessor") else None
        
        self.internet_checker   .stop() if hasattr(self, "internet_checker") else None
        self.visualize          .stop() if hasattr(self, "visualize") else None
        self.logger.info("Main Algos childern stopped")

    def check_childern_complete(self):
        objects = [
            "visualize",
            "text_to_speech",
            "garbage_collector",
            "auto_idle",
            "internet_checker",
            "auth_provider",
            "fps_er",
            "obstacle_avoidance",
            "object_tracker",
            "match_tracker",
            "path_tracker",
            "scene_text_spotter",
            "image_captioner",
            "image_tagger",
            "cloud_preprocessor",
        ]
        self.childern_complete = all([hasattr(self, obj) for obj in objects])
        return self.childern_complete

    ############################## Subcribe Callbacks ##############################
    def frame0_update(self, data):
        self.frame_raw0 = self.bridge.imgmsg_to_cv2(data, "passthrough")      
        self.last_frame0_update = self.get_clock().now()  
        self.auto_idle.update_frame_time()

    def frame1_update(self, data):
        self.frame_raw1 = self.bridge.imgmsg_to_cv2(data, "passthrough")
        self.last_frame1_update = self.get_clock().now()

    def imu_update(self, data: Float64MultiArray):
        self.imu_data = data.data
        acc = self.imu_data[0:3]
        gyro = self.imu_data[3:6]
        self.auto_idle.update_imu(acc, gyro)

    def ori_udpate(self, data: Float64MultiArray):
        self.ori_data = data.data

    def gps_update(self, data: Float64MultiArray):
        self.gps_data = data.data

    def time_diff(self, start_time: Time):
        now = self.get_clock().now()
        return now, (now - start_time).nanoseconds * 1e-9

    ############################## Main Loop ##############################
    def process_frame(self):
        # ========== Check if it is time to process =================
        if not self.childern_complete:
            return
        if self.state != State.PRIMARY_STATE_ACTIVE:
            return
        if self.is_processing:
            return

        main_frame = self.frame_raw0
        # check if frame 1 has been updated recently (< 0.5s)
        second_camera = self.time_diff(self.last_frame1_update)[1] < 2
        if second_camera:
            detect_frame = self.frame_raw1
        else:
            self.frame_raw1 = None  # reset frame 1 to not use it in drawing
            detect_frame = self.frame_raw0
            
        if main_frame is None or (second_camera and detect_frame is None):
            return
        if not isinstance(main_frame, np.ndarray) or (
            second_camera and not isinstance(detect_frame, np.ndarray)
        ):
            return
        
        # ========== Start Processing =================
        start_time = self.get_clock().now()
        self.is_processing = True
        
        lt = self.get_clock().now()
        loop_vis = []
        # ========== Internet Check =================
        is_connected = self.internet_checker.connected()

        # ========== Update Machines with Images and Sensor Data =================
        motion_data = self.auto_idle.get_motion_data()
        is_idle = self.auto_idle    .is_idle(SHORT_TERM_IDLE_TIMEOUT)

        # Update the tracker with the sensor data
        self.obstacle_avoidance .update(main_frame, self.imu_data, is_idle)
        self.object_tracker     .update(detect_frame, self.imu_data, is_idle)
        self.match_tracker      .update(main_frame, motion_data, self.object_detected)
        self.path_tracker       .update(self.imu_data, self.ori_data, self.gps_data)

        lt, diff = self.time_diff(lt)
        loop_vis.append(diff)
        # ========== Scene Text spotting and Other MS Models =================
        # coupled the cloud models with image matching,
        # turn off the cloud models when it is not a keyframe
        if self.config.image_path_following.enable and not self.is_decision_point:
            self.scene_text_spotter .set_enable(False)
            self.image_captioner    .set_enable(False)
            self.image_tagger       .set_enable(False)
        else:
            # all the other cases:
            # 1. not using image matching, do cloud infer
            # 2. using image matching, but this is a keyframe, do cloud infer
            self.scene_text_spotter .set_enable(self.config.cloud.scene_text_spotting)
            self.image_captioner    .set_enable(self.config.cloud.image_captioning)
            self.image_tagger       .set_enable(self.config.cloud.image_tagging)

        if (
            self.config.cloud.scene_text_spotting
            or self.config.cloud.image_captioning
            or self.config.cloud.image_tagging
        ):
            self.cloud_preprocessor .update(detect_frame, self.scene_text_spotter.size, is_idle)
            cloud_image = self.cloud_preprocessor.get_image_bytes()
            self.scene_text_spotter .set_image(cloud_image)
            self.image_captioner    .set_image(cloud_image)
            self.image_tagger       .set_image(cloud_image)
            self.scene_text_spotter .detect(is_idle, is_connected)
            self.image_captioner    .detect(is_idle, is_connected)
            self.image_tagger       .detect(is_idle, is_connected)

        lt, diff = self.time_diff(lt)
        loop_vis.append(diff)
        
        # ========== Signaling =================
        (
            obs_dir,
            obs_dist,
            contours,
            apf_route,
            goal,
            obs_turning_point,
            obstacle_detect,
        ) = self.obstacle_avoidance.get_data()
        # prmask = self.obstacle_avoidance.pr_mask
        # self.pr_debug.publish(self.bridge.cv2_to_imgmsg(prmask, "passthrough")) if prmask is not None else None
        (
            obj_dir,
            obj_dist,
            obj_align,
            object_detected,
        ) = self.object_tracker.get_data()
        (
            matcher_states,
            template_info,
        ) = self.match_tracker.get_data()
        if template_info:
            self.is_decision_point = template_info["is_decision_point"]
        map_data = self.path_tracker.get_data()

        lt, diff = self.time_diff(lt)
        loop_vis.append(diff)
        
        # ========== Publish Algo Data =================
        self.obs_data_pub.publish(
            ObstacleAvoidanceData(
                obs_dir=    float(obs_dir),
                obs_dist=   float(obs_dist),
            )
        ) if obs_dir is not None and obs_dist is not None else None
        self.obj_data_pub.publish(
            ObjectTrackingData(
                obj_dir=    float(obj_dir),
                obj_dist=   float(obj_dist),
                obj_align=  obj_align,
            )
        ) if obj_dir is not None and obj_dist is not None else None
        self.match_data_pub.publish(
            ImageMatchingData(
                match_align=            matcher_states["align"],
                match_finished_trip=    matcher_states["finished_trip"],
                match_micro_adjustment= matcher_states["micro_adjust_dir"],
                kf_turning_dir=         template_info["kf_turning_dir"],
                kf_dist=                template_info["kf_dist"],
            )
        ) if matcher_states and template_info else None
        self.map_data_pub.publish(
            MapPathFollowingData(
                map_bearing_to_path=    map_data["map_bearing_to_path"],
                map_align=              map_data["map_align"],
                map_end_of_path=        map_data["map_end_of_path"],
            )
        ) if map_data else None

        lt, diff = self.time_diff(lt)
        loop_vis.append(diff)

        # ========== VISUALIZATION =================
        text_detected = self.scene_text_spotter .get_objects()
        image_caption = self.image_captioner    .get_objects()
        image_tags = self.image_tagger          .get_objects()
        map_path_debug_info = (
            {
                "lat":                  self.path_tracker.get_cord()[0],
                "lng":                  self.path_tracker.get_cord()[1],
                "map_next_dir":         map_data["map_next_dir"],
                "map_next_dist":        map_data["map_next_dist"],
                "map_bearing_to_path":  map_data["map_bearing_to_path"],
                "map_on_path":          map_data["map_on_path"],
            }
            if map_data
            else {}
        )
        image_matching_debug_info = (
            {
                "match_align":      matcher_states["align"],
                "first_match":      matcher_states["first_match"],
                "match_trip_done":  matcher_states["finished_trip"],
                "match_idx":        matcher_states["template_idx"],
                "n_matches":        matcher_states["n_matches"],
                "micro_adjust_dir": matcher_states["micro_adjust_dir"],
                "keyframe_desc":    template_info["keyframe_desc"],
            }
            if matcher_states and template_info
            else {}
        )
        
        self.visualize.update(
            is_idle,
            self.frame_raw0,
            self.frame_raw1,
            self.fps_er.avg,
            contours,
            apf_route,
            goal,
            obs_turning_point,
            obs_dir,
            obstacle_detect,
            matcher_states,
            sensor_data=(
                self.imu_data + self.ori_data + self.gps_data
                if self.imu_data
                else None
            ),
            loop_vis=self.loop_vis,
            object_detected=object_detected,
            text_detected=text_detected,
            image_caption=image_caption,
            image_tags=image_tags,
            debug_info={
                "obs_dist": obs_dist,
                **map_path_debug_info,
                **image_matching_debug_info,
                "tts_text": self.text_to_speech.text,
            },
        )
        ann0, ann1 = self.visualize.get_annotations()

        lt, diff = self.time_diff(lt)
        loop_vis.append(diff)
        
        # ========== Publish Annotations =================

        self.ann0_pub.publish(
            self.bridge.cv2_to_imgmsg(ann0, "bgr8")
        ) if ann0 is not None else None
        self.ann1_pub.publish(
            self.bridge.cv2_to_imgmsg(ann1, "bgr8")
        ) if ann1 is not None and second_camera else None

        lt, diff = self.time_diff(lt)
        loop_vis.append(diff)

        # ========== Text To Sppech =================
        audio_clue = self.text_to_speech.speak(
            matcher_states,
            template_info,
            obs_dir,
            obs_dist,
            image_caption,
            image_tags,
            text_detected,
        )
        self.tts_text.publish(String(data=audio_clue))
        
        # ========== Exit =================
        self.garbage_collector.collect()
        self.is_processing = False
        self.fps_er.append(1 / ((1/self.config.rate)+ self.time_diff(start_time)[1]))
        self.loop_vis = loop_vis


def main(args=None):
    rclpy.init(args=args)
    node = Algorithms()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        if node.state != State.PRIMARY_STATE_FINALIZED:
            node.stop_childern()
            node.cleanup_childern()
    # algorithms.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()
