import numpy as np
import os
import time
from .utils import *
from algorithms.constants import MAIN_SERVER,MAIN_SERVER_PORT
from algorithms.general.internet_checker import ping_server
from algorithms.general.auth import AuthProvider
from algorithms.general.utils import PeriodicThread
from algorithms.config import Config

class PathTracker:
    def __init__(self, root:str, config: Config, logger, auth_provider:AuthProvider=None):
        self.config = config
        self.root = root
        self.logger = logger
        self.auth_provider = auth_provider

        self.curr_coord = [-1,-1]
        self.hist_coord = []
        self.curr_dir = 0
        self.hist_dir = []
        
        self.next_map_coord = [-1,-1]
        self.next_map_dir = 0.0
        self.next_map_dist = 0.0
        self.bearing_to_path = 0.0
        self.hist_bearing_to_path = []

        self.on_path =False
        self.close_to_path_thres= 8 
        
        self.end_of_path = False

        self.gps_receiving = False       
         
        self.reconfigure(config)
        self.thread = PeriodicThread(0.1, self._get_next_cord_and_dir)
        
    def start(self):
        if self.thread:
            self.thread.start()
    def stop(self):
        if self.thread:
            self.thread.stop()
    def release(self):
        self.stop()                
        
    def reconfigure(self, config: Config):
        self.config = config
        self.get_route()

    def get_cord(self):
        return self.orginal_curr_coord
    
    def get_sensor_dir(self):
        return self.curr_dir
    
    def get_gps_receiving(self):     
        return self.gps_receiving

    def get_route(self):
        if not self.config.map_path_following.enable: return

        if self.config.map_path_following.mock_route:
            if not os.exeists(self.config.map_path_following.mock_route_folder):
                self.logger.error(f'Cannot find the mock route folder {self.config.map_path_following.mock_route_folder}')
                return
            path = get_path_from_txt(self.config.map_path_following.mock_route_folder,3)
            self.path_to_follow = [i[1] for i in path]
        else:
            assert ping_server(MAIN_SERVER, MAIN_SERVER_PORT, timeout=1), 'Main server Not Responsive, Please disable useMapRoute'
            header, device_info = self.auth_provider.get_header_and_device_info()
            self.path_to_follow = get_path_from_map_server(header, device_info)

        self.path_to_follow = [gcs_to_cartesian(i) for i in self.path_to_follow]

    # Update the user's coordinate and others
    def update(self,imu, gps, orientation):
        if not self.config.map_path_following.enable:
            return
        lat, lng = gps
        heading = orientation[2]

        # set the gps receiving flag
        if (abs(lat + 1) >= 10.01 or abs(lng + 1) >= 10.01) and not self.gps_receiving:
            self.gps_receiving = True
        elif (abs(lat ) <= 0.001 or abs(lng ) <= 0.001) and self.gps_receiving:
            self.gps_receiving = False
  
        self.orginal_curr_coord = [lat, lng]      
        self.curr_coord = gcs_to_cartesian(self.orginal_curr_coord)     
        self.curr_dir = heading
        
        if (len(self.hist_coord) >= 1 and dist_from_coord(self.curr_coord, self.hist_coord[-1])>1e-1 
            or len(self.hist_coord) == 0):
            self.hist_coord.append(self.orginal_curr_coord)
            self.hist_dir.append(self.curr_dir)

    def get_data(self):
        if not self.config.map_path_following.enable: return None
        
        align = True if abs(self.bearing_to_path) < 35 else False
        map_next_coord = cartesian_to_gcs(self.next_map_coord)
        return dict(
            map_next_coord = map_next_coord, 
            map_next_dist = self.next_map_dist, 
            map_next_dir = self.next_map_dir, 
            map_bearing_to_path = self.bearing_to_path,
            map_align = align,
            map_on_path = self.on_path, 
            map_end_of_path = self.end_of_path
        )
            
    def _get_next_cord_and_dir(self):
        if not self.config.map_path_following.enable: return
        lt = time.time()
        # get the next coordinate to follow
        # the minimum distance to the next coordinate is 1m
        route_stats = [dist_to_line_between_two_points(i, self.curr_coord) for i in list(zip(self.path_to_follow[:-1], self.path_to_follow[1:]))]
        route_stats = np.array(route_stats,dtype=object)
        
        # get the minimum distance
        # if there are multiple minimum distance, choose the one with the larger index
        duplicate_idx = np.where(np.diff(route_stats[:,0]) == 0)[0]
        min_idx = np.argmin(route_stats[:,0])
        min_idx = min_idx if min_idx not in duplicate_idx else min_idx+1
        
        
        _,b = self.path_to_follow[min_idx],self.path_to_follow[min_idx+1]

        self.next_map_dist = route_stats[min_idx,0]
        self.next_map_dir = route_stats[min_idx,1]
        self.next_map_coord = route_stats[min_idx,2]
        self.on_path = False

        ## pretty much on path, follow the target instead 
        if self.next_map_dist <  self.close_to_path_thres:
            self.next_map_coord = b
            self.next_map_dir = dir_from_coord(self.curr_coord, b)
            self.next_map_dist = dist_from_coord(self.curr_coord, self.next_map_coord)
            self.on_path = True
        
        # if the next coordinate is the last coordinate and the distance is less than 0.5m
        if min_idx+1 == len(self.path_to_follow)-1 and self.next_map_dist < 0.3:
            self.end_of_path = True
        else:
            self.end_of_path = False

        self.bearing_to_path= diff_from_curr_and_target_dir(self.curr_dir, self.next_map_dir)
        self.hist_bearing_to_path.append(self.bearing_to_path)

        # print(f'get_next_cord_and_dir took {time.time()-lt:.3f}s')
    