from math import sqrt, atan2,  pi
import requests
import ast
from pyproj import Transformer
from algorithms.constants import MAIN_SERVER

G2C = Transformer.from_crs(crs_from = "EPSG:4326", crs_to = 'EPSG:32649', always_xy = True)
C2G = Transformer.from_crs(crs_from = "EPSG:32649", crs_to = 'EPSG:4326', always_xy = True)

def coords_interpolate(a, b, t):
    '''Interpolate between two coordinates a and b at time t
    t is a number between 0 and 1
    '''
    lat1, lng1 = a
    lat2, lng2 = b
    lat = lat1 + t * (lat2 - lat1)
    lng = lng1 + t * (lng2 - lng1)
    return lat, lng

def zone(lat, lng):
    '''
    determine the zone of the UTM coordinate
    '''
    if(lat > 0): # Northern Hemisphere
        zone = int(lng/6 + 30)
        EPSG = f'EPSG:326{zone}'

    else: # Southern Hemisphere
        zone = int(lng/6 + 30)
        EPSG = f'EPSG:327{zone}'
    return EPSG

def gcs_to_cartesian(transformer, a):
    '''
    transform from lat, lng to a local cartesian coordinate
    inputs: lat, lng
    outputs: x, y
    '''
    lat = a[0]
    lng = a[1]
    # code = zone(lat, lng)
    # transformer = pyproj.Transformer.from_crs(crs_from = "EPSG:4326", crs_to = code, always_xy = True)
    x, y = G2C.transform(lng, lat) # EPSG:32649
    return x, y
def cartesian_to_gcs(a):
    '''
    transform from local cartesian coordinate to lat, lng
    inputs: x, y
    outputs: lat, lng
    '''
    x = a[0]
    y = a[1]
    # code = zone(0, 0)
    # transformer = pyproj.Transformer.from_crs(crs_from = code, crs_to = "EPSG:4326", always_xy = True)
    lng, lat = C2G.transform(x, y) # EPSG:32649
    return lat, lng

# Get the direction of the user based on their path
# Mimic the function of a compass
################################
#                0
#       315             45
#  270                      90
#       225             135
#           180.1|179.9
################################
def dir_from_coord(a, b):
    '''
    inputs: x,y coord
    outputs: degree
    '''
    x1, y1 = a[0], a[1]
    x2, y2 = b[0], b[1]

    dx = x2 - x1
    dy = y2 - y1

    bearing = atan2(dx, dy) * 180 / pi
    if bearing < 0:
        bearing += 360

    return bearing

def dist_from_coord(a, b):
    '''Return the distance between two points in meters'''

    x1, y1 = a[0], a[1]
    x2, y2 = b[0], b[1]

    dx = x2 - x1
    dy = y2 - y1

    distance = sqrt(dx**2 + dy**2)

    return distance

def closest_coord(a, b, p):
    
    x1, y1 = a[0], a[1]
    x2, y2 = b[0], b[1]
    p1, p2 = p[0], p[1]

    ab =  dist_from_coord(a, b)
    if ab == 0:  # Line segment has zero length
        return a

    t = ((p1 - x1) * (x2 - x1) + (p2 - y1) * (y2 - y1)) / ab # AP dot AB, divided by squared length of AB
    t = max(0, min(1, t))  # Clamp t between 0 and 1
    

    closestX = x1 + t * (x2 - x1)
    closestY = y1 + t * (y2 - y1)

    return [closestX,closestY]


def dist_to_line_between_two_points(line, point):
    '''Return the distance between a line and a point
    line is a list of two coordinates
    point is a coordinate
    return dxa: distance in meters
        bearing: bearing in radians
        next_coord: the coordinate of the closest point on the line
    '''
    [x1, y1] = line[0]
    [x2, y2] = line[1]
    [x3, y3] = point

    a, b, me = [x1, y1], [x2, y2], [x3, y3]

    next_coord = closest_coord(a, b, me)

    dxa =  dist_from_coord(me, next_coord)
    bearing =  dir_from_coord(me, next_coord)

    if bearing < 0:
        bearing += 360
    
    return dxa, bearing, next_coord 

def diff_from_curr_and_target_dir(curr_dir, target_dir):
    # curr_dir and target_dir are in degrees
    # top is 0, left dir is in negative, right dir is in positive, bottom is 180
    diff = target_dir - curr_dir
    if diff > 180:
        diff = diff - 2 * 180
    elif diff < -180:
        diff = diff + 2 * 180
    return diff

def get_path_from_map_server(header , device_info):
    '''
    Fetch path from map server
    header: dict
    device_info: dict
    '''
    try:
        all_coords  =[]
        path_to_follow = requests.post(
            f'{MAIN_SERVER}/ghopper/get_route',
            headers=header,
            params={
                'starting_lat': device_info['starting_lat'],
                'starting_lon': device_info['starting_lon'],
                'ending_lat': device_info['ending_lat'],
                'ending_lon': device_info['ending_lon'],
            },
            timeout=10,
        ).json()
        # TODO: Reflect multiple paths and mutliple segments with segment type (e.g. bus, foot, etc.)
        for name, segments in path_to_follow.items():
            for segment in segments:
                coords = segment.get('Path', [])                
                for coord in coords:
                    all_coords.append(coord)
            break
    except Exception as e:
        print(e)
        all_coords = None
    return all_coords

# To load a specified sequence of hard-coded coordinates and return a list
# Start a sequence of coordinate with the line 'plan=X' where X is an integer
# each step is [dir, lat, lng, timestamp]
def get_path_from_txt(filename, plan):
    assert type(plan) == int

    curr_plan = -1

    try:
        f = open(filename, 'r')
    except FileNotFoundError:
        return []
    

    steps = []
    for line in f:
        if not line.strip():
            continue
        if line[0] == '#':
            continue
        
        if line[0:4] == 'plan':
            line = line.strip().split('=')
            curr_plan = int(line[1].strip())
            continue

        if curr_plan != plan:
            continue
        line = line.strip()
        data = ast.literal_eval(line)
        dir = data[0]
        coords = (data[1],data[2])
        timestp=data[3]
        steps.append((dir,coords,timestp ))
    return steps