import numpy as np

from algorithms.constants import HEIGHT
from .sort import Sort
from .utils import *
from algorithms.config import Config
from algorithms.general.utils import PeriodicThread

from enum import Enum

class ErrorCodes(Enum):
    SUCCESS = 0
    NO_MODEL = 1
    NO_IMAGE = 2
    IDLE = 3
    

class ObjectTracker:
    def __init__(self, root, config: Config, logger):
        self.root = root
        self.config = config
        self.reconfigure(self.config)
        self.init_model()
        self.logger = logger
        self.thread = PeriodicThread(0.05, self.compute)
                
    def start(self):
        if self.thread:
            self.thread.start()        
            
    def stop(self):
        if self.thread:
            self.thread.stop()
    
    def release(self):
        self.stop()
        if self.object_detector is not None:
            self.object_detector.destroy()
            
    def init_model(self):
        if self.config.engine_type.lower() == "onnx":
            from .object_detection.object_detection_onnx import (
                ObjectDetectionONNX as ObjectDetection,
            )
        elif self.config.engine_type.lower() == "bpu":
            from .object_detection.object_detection_bpu import (
                ObjectDetectionBPU as ObjectDetection,
            )
        self.object_detector = ObjectDetection(self.root, self.config)
            
    def reconfigure(self,config:Config):
        self.image=None
        self.curr_dir = 0
        self.next_obj_dir = 0
        self.next_obj_dist = HEIGHT
    
        self.frame= 0
        self.object_expire_time = 20
        self.objects = []

        self.is_idle = False
        self.mot_tracker = Sort(max_age=1, 
                            min_hits=3,
                            iou_threshold=0.3) 
        self.return_code = ErrorCodes.SUCCESS
            
    # Update the user's coordinate and others
    def update(self, image,sensor_data,is_idle,):
        if not self.config.object_tracking.enable: return

        # self.curr_dir = sensor_data[SENSOR_MAP['dir']]
        self.is_idle = is_idle
        self.image = image

    def compute(self):      
        if not self.config.object_tracking.enable:
            return
        
        if self.return_code not in [ErrorCodes.SUCCESS]:
            self.logger.debug(f"Object Detection Failed: {self.return_code}")
        if self.image is None:
            self.return_code = ErrorCodes.NO_IMAGE
            return
        if self.object_detector is None:
            self.return_code = ErrorCodes.NO_MODEL
            return
        if self.is_idle:
            self.return_code = ErrorCodes.IDLE            
            return
        
        self.return_code = ErrorCodes.SUCCESS
        image = self.object_detector.preprocess(self.image)
        objects = self.object_detector.detect(image)
        
        dets = []
        in_cls = []
        for box in objects:
            if box['id'] == 'car':
                self.next_obj_dist = HEIGHT - box['bottom']
        
            dets.append([box['left'],box['top'],box['right'],box['bottom'],box['score']])
            in_cls.append(box['id'])

            box['cls'] = box['id']
            box['id'] = 'N/A'

        if not self.config.object_tracking.use_tracking: 
            self.objects = objects
            return

        if len(dets) == 0:
            dets = np.empty((0, 5))
        dets = np.array(dets)
        trackers,object_cls = self.mot_tracker.update(dets,in_cls)
        final_objects = []
        for d ,c in zip(trackers,object_cls):    
            final_objects.append(
                {
                    'left':int(d[0]),
                    'top':int(d[1]),
                    'right':int(d[2]),
                    'bottom':int(d[3]),
                    'cls':c, 
                    'id':d[4]
                }
            )
        self.objects = final_objects        
        
    def get_data(self):
        return 0, self.get_dist(), False, self.objects

    # TODO: get the obj dist from image matching
    def get_dist(self):
        return self.next_obj_dist
    