
import queue
from enum import Enum

from algorithms.general.utils import PeriodicThread
from algorithms.config import Config
from algorithms.constants import MAX_TURN_BUFFER

from .route import Route
from .goal import Goal
from .obstacle_detect import ObstacleDetect
from .utils import get_obs_direction
from .binary import get_binary_e2e

class ErrorCodes(Enum):
    SUCCESS = 0
    NO_MODEL = 1
    NO_IMAGE = 2
    IDLE = 3
    PREPROCESS_FAILED = 4
    SEGMENTATION_FAILED = 5
    

class ObsatcleAvoidance:
    def __init__(self, root, config: Config, logger):
        self.root = root
        self.config = config
        self.reconfigure(self.config)
        self.init_model()
        self.logger = logger
        self.thread = PeriodicThread(0.05, self.compute)
        
    def start(self):
        if self.thread and not self.thread.running:
            self.thread.start()        
            
    def stop(self):
        if self.thread and self.thread.running:
            self.thread.stop()
    
    def release(self):
        self.stop()
        if self.seg_model is not None:
            self.seg_model.destroy()
    
    def init_model(self):
        if self.config.engine_type.lower() == "onnx":
            from .segmentation.segmentation_onnx import (
                SegmentationONNX as Segmentation,
            )
        elif self.config.engine_type.lower() == "bpu":
            from .segmentation.segmentation_bpu import (
                SegmentationBPU as Segmentation,
            )
        self.seg_model = Segmentation(self.root, self.config)

    def reconfigure(self, node_config: Config):
        self.image=None
        self.obstacle_detect = ObstacleDetect(node_config)
        self.route = Route(node_config)
        self.goal_getter = Goal()
        self.turn_queue = queue.Queue(MAX_TURN_BUFFER)
        self.binary_mask = None
        self.contours = None
        self.runaway_goal = None
        self.apf_route = None
        self.turning_point = None
        self.obs_dir = None
        self.is_idle = False
        self.obs_dist = None
        self.obs_turning_point = None
        self.sensor_data = None
        self.return_code = ErrorCodes.SUCCESS

    def update(self, image, sensor_data,is_idle):
        if not self.config.obstacle_avoidance.enable: return
        
        # self.sensor_data = sensor_data
        self.is_idle = is_idle
        self.image = image
        
    def compute(self):
        if not self.config.obstacle_avoidance.enable:
            return
        
        if self.return_code not in [ErrorCodes.SUCCESS]:
            self.logger.debug(f"Obstacle Avoidance Failed: {self.return_code}")
        if self.image is None:
            self.return_code = ErrorCodes.NO_IMAGE
            return        
        if self.seg_model is None:
            self.return_code = ErrorCodes.NO_MODEL
            return
        if self.is_idle:
            self.return_code = ErrorCodes.IDLE            
            return
        
        self.return_code = ErrorCodes.SUCCESS
        image, valid = self.seg_model.preprocess(self.image)
        if not valid:
            self.return_code = ErrorCodes.PREPROCESS_FAILED
            
        pr_mask, valid = self.seg_model.infer(image)
        if not valid:
            self.return_code = ErrorCodes.SEGMENTATION_FAILED

        self.binary_mask, self.contours = get_binary_e2e(pr_mask)
        self.runaway_goal = self.goal_getter.get_goal(self.binary_mask)
        self.apf_route, self.obs_turning_point = self.route.get_route(self.binary_mask, self.runaway_goal)
        self.obs_dir = get_obs_direction(self.obs_turning_point, self.turn_queue)
        self.obs_dist = self.obstacle_detect.get_obstacle(self.binary_mask,self.contours)
        

    def get_data(self):
        return (
            self.obs_dir, self.obs_dist,
            self.contours, self.apf_route, self.runaway_goal,
            self.obs_turning_point,
            self.obstacle_detect,
        )