import cv2
import numpy as np
import time
from math import exp
from algorithms.constants import HEIGHT,  WIDTH
from .line import Line
from algorithms.general.utils import LimitedQ

# Obstacle detection based on edge detection
class ObstacleDetect:
    def __init__(self, config):
        self.config = config

        self.LEFT_DEFAULT_LINE = (
            int(0.375 * WIDTH), 
            HEIGHT - 1, 
            int(0.425 * WIDTH), 
            0
        )
        self.RIGHT_DEFAULT_LINE = (
            int(0.625 * WIDTH),
            HEIGHT - 1,
            int(0.575 * WIDTH),
            0,
        )
        self.left_tangent = Line(self.LEFT_DEFAULT_LINE)
        self.right_tangent = Line(self.RIGHT_DEFAULT_LINE)
        self.hori_tangent = Line()

        self.obs_dist = LimitedQ(10)
        self.bottom_y = HEIGHT - 1

        self.vert_line_mask = self.vert_line_extract_region()
        self.hori_line_mask = self.hori_line_extract_region()

    def extract_contour_mask(self, contours):
        sorted_binary_cnts = sorted(contours, key=cv2.contourArea, reverse=True)

        contour_mask = np.zeros((HEIGHT, WIDTH), dtype=np.uint8)
        for cnt in sorted_binary_cnts:
            contour_mask = cv2.drawContours(contour_mask, [cnt], -1, 1, thickness=2)
        # contour_mask = cv2.GaussianBlur(contour_mask, (5, 5), 0)
        return contour_mask

    def vert_line_extract_region(self):
        # Region to consider the vertical lines
        vert_line_mask = np.zeros((HEIGHT, WIDTH), dtype=np.uint8)

        # Region to find the left vertical line
        left_region = [
            [int(0.2 * WIDTH), HEIGHT - 1],
            [int(0.25 * WIDTH), 0],
            [int(0.475 * WIDTH), 0],
            [int(0.475 * WIDTH), HEIGHT - 1],
        ]
        vert_line_mask = cv2.fillPoly(
            vert_line_mask, pts=np.array([left_region]), color=1
        )

        # Region to find the right vertical line
        right_region = [
            [WIDTH - 1 - left_region[0][0], HEIGHT - 1],
            [WIDTH - 1 - left_region[1][0], 0],
            [WIDTH - 1 - left_region[2][0], 0],
            [WIDTH - 1 - left_region[3][0], HEIGHT - 1],
        ]
        vert_line_mask = cv2.fillPoly(
            vert_line_mask, pts=np.array([right_region]), color=1
        )
        return vert_line_mask

    def hori_line_extract_region(self):
        hori_line_mask = np.zeros((HEIGHT, WIDTH), dtype=np.uint8)

        region = [
            [int(0.1 * WIDTH), int(0.9*HEIGHT)],
            [int(0.1 * WIDTH), int(0.1 * HEIGHT)],
            [int(0.9 * WIDTH), int(0.1 * HEIGHT)],
            [int(0.9 * WIDTH), int( 0.9*HEIGHT)],
        ]
        hori_line_mask = cv2.fillPoly(hori_line_mask, pts=np.array([region]), color=1)
        return hori_line_mask

    def extract_vert_line(self, vert_line_mask):
        lines = cv2.HoughLinesP(
            vert_line_mask,  # Input edge image
            1,  # Distance resolution in pixels
            np.pi / 180,  # Angle resolution in radians
            threshold=50,  # Min number of votes for valid line
            minLineLength=40,  # Min allowed length of line
            maxLineGap=15,  # Max allowed gap between line for joining them
        )
        left_x_list = []
        right_x_list = []
        if lines is None:
            lines = []
        for points in lines:
            line = Line(points[0])

            if line.slope < 1 and line.slope > -1:
                continue

            if line.slope > 0 and line.x1 < (WIDTH // 2 - 1):
                continue
            elif line.slope < 0 and line.x1 > (WIDTH // 2 - 1):
                continue

            if line.x1 < WIDTH // 2:
                # left_lines_list.append([(x1,y1),(x2,y2)])
                left_x_list.append(((line.x1 + line.x2) // 2))
            elif line.x1 >= WIDTH // 2:
                # right_lines_list.append([(x1,y1),(x2,y2)])
                right_x_list.append(((line.x1 + line.x2) // 2))

        if len(left_x_list) > 0:
            xcoord = int(np.percentile(left_x_list, 90))
            self.left_tangent.update_coord((xcoord, 0, xcoord, HEIGHT - 1))
        else:
            self.left_tangent.update_coord(self.LEFT_DEFAULT_LINE)

        if len(right_x_list) > 0:
            xcoord = int(np.percentile(right_x_list, 10))
            self.right_tangent.update_coord((xcoord, 0, xcoord, HEIGHT - 1))
        else:
            self.right_tangent.update_coord(self.RIGHT_DEFAULT_LINE)

        return True

    def extract_hori_line(self, hori_line_mask):
        lt = time.time()
        lines = cv2.HoughLinesP(
            hori_line_mask,  # Input edge image
            5,  # Distance resolution in pixels
            np.pi / 180,  # Angle resolution in radians
            threshold=50,  # Min number of votes for valid line
            minLineLength=40,  # Min allowed length of line
            maxLineGap=15,  # Max allowed gap between line for joining them
        )

        hori_y_list = []
        if lines is None:
            lines = []

        for points in lines:
            line = Line(points[0])
            if not (self.left_tangent.valid or self.right_tangent.valid):
                continue

            if ((line.x1 + line.x2) // 2) < self.left_tangent.x1 or \
                ((line.x1 + line.x2) // 2) > self.right_tangent.x1:
                continue
            hori_y_list.append(((line.y1 + line.y2) // 2))

        if len(hori_y_list) > 0:
            ycoord = int(np.percentile(hori_y_list, 90))
            self.hori_tangent.update_coord((0, ycoord, WIDTH - 1, ycoord))
        else:
            self.hori_tangent.update_coord()

        return True

    def get_obstacle(self,binary_mask, contours):
        if not self.config.obstacle_avoidance.use_obstacle_detect:
            return HEIGHT

        self.obs_dist.append(HEIGHT)
        # Regions to extract the small vertical lines
        vert_line_regions = self.vert_line_mask
        hori_line_regions = self.hori_line_mask

        # Get the mask of the binary mask's contour (outline only)
        contour_mask = self.extract_contour_mask(contours)
        start_point = (WIDTH // 2, HEIGHT - 1)
        if np.count_nonzero(contour_mask) < 10:
            # the contours mask is all black, which means there is obstacle immediately in front
            self.obs_dist.append(0)
            return self.obs_dist.avg
        # if binary_mask[start_point[1], start_point[0]] == 1: 
        #     # the pixel at the start point is black, which means there is obstacle immediately in front
        #     return 0
        # Restrict the mask to a specific region (outlines within the region)
        vert_line_mask = np.bitwise_and(contour_mask, vert_line_regions)
        hori_line_mask = np.bitwise_and(contour_mask, hori_line_regions)
        self.extract_vert_line(vert_line_mask)
        self.extract_hori_line(hori_line_mask)       

        bottom_y = HEIGHT - 1

        self.bottom_y = bottom_y
        if (
            self.left_tangent.valid
            and self.right_tangent.valid
            and self.hori_tangent.valid
        ):
            # self.obs_dist = HEIGHT - self.hori_tangent.y1
            self.obs_dist.append(self.bottom_y - self.hori_tangent.y1)
            # if self.hori_tangent.y1 > (0.7 * HEIGHT):
            #     self.obs_found = True
        if not self.hori_tangent.valid:
            self.obs_dist.append(0)

        return self.obs_dist.avg
