import numpy as np
import time
from scipy.ndimage.morphology import distance_transform_edt
from algorithms.constants import HEIGHT, WIDTH
from .utils import pl_dist

DEFUALT_GOAL = [WIDTH // 2, 0]

M_BETA = 0.5

R_D2 = lambda x: (x / 100.0) + 1
R_D0 = 1.35  # increase this to make the replusion field stronger
R_NU = 3500

A_XI = 1 / 200  # decrease this to make the attraction field stronger

SPEED = 10  # increase this to make the APF route propagation faster, but produce more oscillation


class Route:
    def __init__(self, config):
        self.config = config
        self.prev_goals = np.array([])
        self.prev_time = time.time()
        self.turning_idx = 0

    def get_route(self, binary_mask, runaway_goal):
        # Get new starting coordinate if the contour does
        # not cover the original starting coordinate
        start = [WIDTH // 2, HEIGHT - 2]  # 160, 190
        final_goal = DEFUALT_GOAL
        x_mask = np.where(binary_mask[:, start[0]] == 0)[0]

        # Split the array into parts, where the difference between the elements is 1, and take the last part
        x_mask = np.split(x_mask, np.where(np.diff(x_mask) != 1)[0] + 1)[-1]

        # If the contour does not cover the original starting coordinate
        if len(x_mask) == 0:
            final_goal = runaway_goal
        # If the contour have an obstacle in the middle lower part of the image
        elif np.min(np.array(x_mask)) > HEIGHT * 0.8 or np.max(np.array(x_mask)) < HEIGHT-2:
            final_goal = runaway_goal
        # print(np.array(x_mask))
        start = np.array(start)
        final_goal = np.array(final_goal)
    
        if self.config.obstacle_avoidance.use_apf_route:
            gradient_map = self.get_gradient_map(binary_mask, final_goal)
            route, turning_point = self.apf_route(
                start,
                final_goal,
                runaway_goal,
                gradient_map,
                iter=10,
                plateau_length=3,
                plateau_count_thres=5,
            )

        else:
            arrow_max_length = HEIGHT // 2
            dist_to_goal = np.clip(np.linalg.norm(start - final_goal), 0, arrow_max_length)
            # calculate the end point of the arrow, with the same angle as the line start-goal
            ratio = dist_to_goal / np.linalg.norm(start - final_goal)
            ending_point = start + ratio * (final_goal - start)
            start = start.astype(int).tolist()
            ending_point = ending_point.astype(int).tolist()
                
            route = np.vstack([start,ending_point])
            route = route.reshape((-1, 1, 2))
            turning_point =  final_goal
        return route, turning_point

    def get_gradient_map(self, binary_mask, goal):
        ### get repulsive potential
        d = distance_transform_edt(binary_mask == 0)
        d2 = R_D2(d)  # Rescale and transform distances
        repulsive = R_NU * ((1.0 / d2 - 1 / R_D0) ** 2)
        repulsive[d2 > R_D0] = 0

        # TODO: increase repulsive / increase attractve

        ### get attractive potential
        nrows, ncols = binary_mask.shape
        [x, y] = np.meshgrid(np.arange(ncols), np.arange(nrows))
        attractive = A_XI * ((x - goal[0]) ** 2 + (y - goal[1]) ** 2)

        ### total potential
        f = attractive + repulsive
        # Gradient planner
        [gy, gx] = np.gradient(-f)
        return gy, gx

    def apf_route(
        self,
        start,
        final_goal,
        runaway_goal,
        gradient_map,
        iter,
        plateau_length,
        plateau_count_thres,
    ):
        gy, gx = gradient_map

        ## algo init
        highest_straight_point = start
        plateau_count = 0
        route = np.vstack([np.array(start)])

        m = 0.0
        plateau_running_mean = 0.0
        ## begin algo loop
        for i in range(iter):
            current_point = route[-1]

            # distance with goal check
            if (
                sum(abs(current_point - final_goal)) < 5.0
                or current_point[1] < final_goal[1] + 3
            ):
                # print('Reached the goal !')
                break

            ix = np.clip(int(round(current_point[1])), 0, HEIGHT - 1)
            iy = np.clip(int(round(current_point[0])), 0, WIDTH - 1)

            ##  gradient descent with momentum
            # obtain velocity
            vx, vy = gx[ix, iy], gy[ix, iy]
            denorm = np.linalg.norm([vx, vy])
            dt = 0 if denorm == 0 else 1 / denorm
            scale = SPEED
            dt *= scale

            # momentum
            m = M_BETA * m + (1 - M_BETA) * dt * np.array([vx, vy])
            next_point = current_point + m

            plateau_running_mean += route[-1] / plateau_length
            ## check if the point has plateaued (keep going in circle in small increment)
            if len(route) > plateau_length:
                plateau_running_mean -= route[-plateau_length - 1] / plateau_length
                s = sum(abs(plateau_running_mean - next_point))
                if s < 1.0 * scale:
                    plateau_count += 1
            if plateau_count > plateau_count_thres:
                break
            # go sideways if the repulsive field is right in front
            # and next_point is out of the picture on the downside
            if next_point[1] > HEIGHT - 1:
                if runaway_goal[0] - WIDTH // 2 > 0:
                    # goal on the right
                    next_point = current_point + np.array([vy, 0])
                else:
                    # goal on the left
                    next_point = current_point - np.array([vy, 0])

            # record the highest point on the straight line
            if abs(next_point[0] - WIDTH // 2) < 2:
                highest_straight_point = next_point

            route = np.vstack([route, next_point])
        ## get turning point
        turning_point = self.get_turning_point(
            route, final_goal, highest_straight_point
        )

        route = np.array(route, np.int32)
        route = route.reshape((-1, 1, 2))
        return route, turning_point

    def get_turning_point(self, route, goal, highest_straight_point):
        ## record max pl dist
        # if next_point is high enough
        # and have the max_dist out of all points with the line start-goal

        dists = [pl_dist([route[0], goal], point) for point in route]
        max_dist_idx = np.argmax(dists)
        idx_diff = clip(max_dist_idx - self.turning_idx, 0, len(route) - 1)
        # idx_diff to large, add to prevent sudden changes for the turning point
        if abs(idx_diff) > 4:
            self.turning_idx += idx_diff
        else:
            self.turning_idx = max_dist_idx

        # if the highest point on the straight line is too low, use the turning point
        if highest_straight_point[1] < 0.8 * HEIGHT:
            turning_point = highest_straight_point.astype(int).tolist()
        else:
            turning_point = route[max_dist_idx].astype(int).tolist()

        return turning_point
