#!/usr/bin/env python

'''
Lucas-Kanade tracker
====================
Lucas-Kanade sparse optical flow demo. Uses goodFeaturesToTrack
for track initialization and back-tracking for match verification
between frames.
'''


import numpy as np
import cv2
from algorithms.constants import HEIGHT,WIDTH
from algorithms.general.utils import LimitedQ

def safe_divide(a, b):
    return a / b if b else 0.0

# params for ShiTomasi corner detection
feature_params = dict( maxCorners = 100,
                       qualityLevel = 0.3,
                       minDistance = 3,
                       blockSize = 7,)

# Parameters for lucas kanade optical flow
lk_params = dict( winSize  = (15,15),
                  maxLevel = 2,
                  criteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, 10, 0.03))

class OpticalFlow:
    def __init__(self):
        self.running_optical_flow = LimitedQ(10)
        self.img = None
        self.optical_flow =10.0

        self.prev_gray = None
        self.prev_gray_p =None


    def get_optical_flow(self):
        return self.optical_flow

    def push_frame(self,frame):    
        frame = cv2.resize(frame, (WIDTH, HEIGHT))    
        if self.prev_gray is None or self.prev_gray_p is None:
            self.prev_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            p = cv2.goodFeaturesToTrack(self.prev_gray,  **feature_params)
            self.prev_gray_p = p
            return
        frame_gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        # calculate optical flow
        p1, _, _ = cv2.calcOpticalFlowPyrLK(self.prev_gray, frame_gray, self.prev_gray_p, None, **lk_params)
        
        good_new = p1
        good_old = self.prev_gray_p
        
        distance = 0
        # draw the tracks
        for i,(new,old) in enumerate(zip(good_new,good_old)):
            a,b = new.ravel()
            c,d = old.ravel()            
            distance +=np.sum(np.abs(a-c) +  np.abs(b-d))

        self.running_optical_flow.append(safe_divide(distance,i))
        self.optical_flow = self.running_optical_flow.mean()
        
        self.prev_gray = frame_gray
        p = cv2.goodFeaturesToTrack(frame_gray,  **feature_params)
        self.prev_gray_p = p
